from __future__ import annotations


from cnl2asp.specification.attribute_component import AttributeOrigin, is_same_origin
from cnl2asp.utility.utility import Utility
from .asp_attribute import ASPAttribute
from .asp_element import ASPElement


class ASPAtom(ASPElement):
    def __init__(self, name: str, attributes: list[ASPAttribute], negated: bool = False,
                 is_before: bool = False, is_after: bool = False, is_initial: bool = False, is_final: bool = False):
        self.name = name
        self.attributes = attributes
        self.negated = negated
        self.is_before = is_before
        self.is_after = is_after
        self.is_initial = is_initial
        self.is_final = is_final

    def set_attributes_value(self, attributes: list[ASPAttribute]):
        for attribute in attributes:
            # get the list of matching attributes
            atom_attributes = self.get_attributes_list(attribute.name, attribute.origin)
            for atom_attribute in atom_attributes:
                # update only the first occurrence
                if atom_attribute.is_null():
                    atom_attribute.set_value(attribute.get_value())
                    break

    def get_atom_list(self) -> list[ASPAtom]:
        return [self]

    def is_null(self):
        for attribute in self.attributes:
            if not attribute.is_null():
                return False
        return True

    def has_attribute(self, attribute: ASPAttribute):
        if attribute in self.attributes:
            return True
        return False

    def get_attributes_list(self, name: str, origin: AttributeOrigin = None) -> list[ASPAttribute]:
        if origin:
            return self.get_attributes_list_by_name_and_origin(name, origin)
        else:
            return self.get_attributes_list_by_name(name)

    def get_attributes_list_by_name(self, name: str) -> list[ASPAttribute]:
        attributes = []
        for attribute in self.attributes:
            if attribute.name == name:
                attributes.append(attribute)
        return attributes

    def get_attributes_list_by_name_and_origin(self, name: str, origin: AttributeOrigin = None) -> list[ASPAttribute]:
        attributes = []
        for attribute in self.get_attributes_list_by_name(name):
            if is_same_origin(attribute.origin, origin):
                attributes.append(attribute)
        return attributes

    def __str__(self) -> str:
        if not self.name:
            return ''
        string = ''
        string += 'not ' if self.negated else ''
        string += '\'' if self.is_before else ''
        string += '_' if self.is_initial else ''
        string += '__' if self.is_final else ''
        string += f'{self.name}'
        string += '\'' if self.is_after else ''
        string += '('
        if Utility.PRINT_WITH_FUNCTIONS:
            visited = []
            for attribute1 in self.attributes:
                if attribute1 in visited:
                    continue
                visited.append(attribute1)
                if attribute1.origin and attribute1.origin.name != self.name:
                    tmp_atom = ASPAtom(attribute1.origin.name, [ASPAttribute(attribute1.name, attribute1.get_value(),
                                                                             attribute1.origin.origin)])
                    for attribute2 in self.attributes:
                        if attribute2 in visited:
                            continue
                        if attribute2.origin and attribute1.origin.name == attribute2.origin.name:
                            visited.append(attribute2)
                            tmp_atom.attributes.append(ASPAttribute(attribute2.name, attribute2.get_value(),
                                                                    attribute2.origin.origin))
                    string += str(tmp_atom) + ','
                else:
                    string += str(attribute1) + ','
            if string[-1] == ',':
                string = string[0:-1]
            return string + ')'
        else:
            return string + f'{",".join([str(x) for x in self.attributes])})'

    def __eq__(self, other):
        if not isinstance(other, ASPAtom):
            return False
        return self.name == other.name and self.attributes == other.attributes and self.is_initial == other.is_initial and self.is_after == other.is_after and self.is_before == other.is_before

    def __repr__(self):
        return str(self)



from uuid import uuid4
import inflect

class Utility:
    NULL_VALUE = '_'
    ASP_NULL_VALUE = '_'
    PRINT_WITH_FUNCTIONS = False
    DEFAULT_ATTRIBUTE = 'id'
    LOCKED_KEYWORDS = ["is", "identified", "and", "equal", "to", "are", "has", "be", "have", "by", "with", "in",
                       "exactly", "at", "most", "least", "any", "every", "more", "less", "greater", "after", "highest",
                       "lowest", "smallest", "biggest", "sum", "difference", "product", "division", "where", "between",
                       "whenever", "such", "that", "there", "than", "also", "then", "required", "prohibited", "or"]
    AUTO_ENTITY_LINK = True

    @staticmethod
    def create_unique_identifier():
        return 'x_' + str(uuid4()).replace('-', '_')

    @staticmethod
    def get_singular(string: str) -> str:
        singular = inflect.engine().singular_noun(string)
        return singular if singular else string


from cnl2asp.parser.proposition_builder import PropositionBuilder
from cnl2asp.specification.attribute_component import ValueComponent, AttributeComponent
from cnl2asp.specification.entity_component import EntityComponent
from cnl2asp.specification.operation_component import OperationComponent, Operators
from cnl2asp.specification.problem import Problem
from cnl2asp.specification.proposition import NewKnowledgeComponent
from cnl2asp.specification.signaturemanager import SignatureManager


class Command:
    def execute(self):
        return


class SubstituteVariable(Command):
    def __init__(self, proposition: PropositionBuilder, variable: str, values: list[ValueComponent]):
        self.proposition = proposition
        self.variable = variable
        self.values = values

    def execute(self):
        new_propositions = []
        for proposition in self.proposition.get_propositions():
            for value in self.values:
                derived_proposition = proposition.copy()
                derived_proposition.add_requisite([OperationComponent(Operators.EQUALITY,
                                                                      ValueComponent(self.variable),
                                                                      value)])
                new_propositions.append(derived_proposition)
        self.proposition._original_rule = new_propositions[0]
        self.proposition._derived_rules = new_propositions[1:]


class RespectivelySubstituteVariable(Command):
    def __init__(self, proposition: PropositionBuilder, variable: str, values: list[ValueComponent]):
        self.proposition = proposition
        self.variable = variable
        self.values = values

    def execute(self):
        for proposition, value in zip(self.proposition.get_propositions(), self.values):
                proposition.add_requisite([OperationComponent(Operators.EQUALITY,
                                                                      ValueComponent(self.variable),
                                                                      value)])


class DurationClause(Command):
    def __init__(self, proposition: PropositionBuilder, new_knowledge: NewKnowledgeComponent,
                 duration_value: ValueComponent, attribute_name: str):
        self.proposition = proposition
        self.new_knowledge = new_knowledge
        self.duration_value = duration_value
        self.attribute_name = attribute_name

    def execute(self):
        self.proposition.duration_clause(self.new_knowledge, self.duration_value, self.attribute_name)


class CreateSignature(Command):
    def __init__(self, proposition: PropositionBuilder, entity: EntityComponent):
        self.proposition = proposition
        self.entity = entity

    def execute(self):
        try:
            signature = SignatureManager.get_signature(self.entity.get_name())
        except:
            signature = self.proposition.create_new_signature(self.entity)
            SignatureManager.add_signature(signature)
        signature.set_attributes_value(self.entity.get_keys_and_attributes())
        self.entity.keys = signature.keys
        self.entity.attributes = signature.attributes



from enum import Enum

from cnl2asp.exception.cnl2asp_exceptions import EntityNotFound, AttributeNotFound
from cnl2asp.converter.converter_interface import Converter, NewKnowledgeConverter, ConditionConverter, \
    RequisiteConverter, \
    CardinalityConverter, PropositionConverter
from cnl2asp.utility.utility import Utility
from .attribute_component import AttributeComponent, ValueComponent, AttributeOrigin
from .component import Component
from .entity_component import EntityComponent
from .relation_component import RelationComponent


class PROPOSITION_TYPE(Enum):
    CONSTRAINT = 0
    REQUIREMENT = 1
    PREFERENCE = 2


class AggregateOfComponents(Component):
    def __init__(self, components: list[Component]):
        self.components = components if components else []

    def add_components(self, components: list[Component]):
        self.components += components

    def convert(self, converter: Converter):
        pass

    def copy(self):
        return [component.copy() for component in self.components]

    def get_entities(self) -> list[EntityComponent]:
        entities = []
        for entity in self.components:
            entities += entity.get_entities()
        return entities

    def get_entities_to_link_with_new_knowledge(self) -> list[EntityComponent]:
        entities = []
        for entity in self.components:
            entities += entity.get_entities_to_link_with_new_knowledge()
        return entities


class ConditionComponent(AggregateOfComponents):
    def __init__(self, condition: list[Component] = []):
        super().__init__(condition)

    def convert(self, converter: Converter) -> ConditionConverter:
        return converter.convert_condition(self)

    def copy(self):
        return ConditionComponent(super().copy())


class NewKnowledgeComponent(Component):
    def __init__(self, new_knowledge: EntityComponent, condition: ConditionComponent = None,
                 subject=None, auxiliary_verb=None, objects=None):
        self.new_entity = new_knowledge
        self.condition = condition if condition else ConditionComponent([])
        self.subject = subject
        self.auxiliary_verb = auxiliary_verb
        self.objects = [x.copy() for x in objects] if objects else None

    def convert(self, converter: Converter) -> NewKnowledgeConverter:
        return converter.convert_new_knowledge(self)

    def copy(self):
        return NewKnowledgeComponent(self.new_entity.copy(), self.condition.copy(), self.subject, self.auxiliary_verb, self.objects)

    def get_entities(self) -> list[EntityComponent]:
        entities = [self.new_entity]
        for entity in self.condition.get_entities():
            entities.append(entity)
        return entities

    def get_entities_to_link_with_new_knowledge(self) -> list[EntityComponent]:
        entities = [self.new_entity]
        for entity in self.condition.get_entities_to_link_with_new_knowledge():
            entities.append(entity)
        return entities

class RequisiteComponent(AggregateOfComponents):
    def __init__(self, requisite: list[Component]):
        super().__init__(requisite)

    def convert(self, converter: Converter) -> RequisiteConverter:
        return converter.convert_requisite(self)

    def copy(self):
        return RequisiteComponent(super().copy())


class CardinalityComponent(Component):
    def __init__(self, lower_bound: int | None, upper_bound: int | None):
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def convert(self, converter: Converter) -> CardinalityConverter:
        return converter.convert_cardinality(self)

    def copy(self):
        return CardinalityComponent(self.lower_bound, self.upper_bound)


class Proposition(Component):
    # Data structure for representing the sentences
    def __init__(self, new_knowledge: list[NewKnowledgeComponent] = None,
                 cardinality: CardinalityComponent = None, requisite: RequisiteComponent = None,
                 relations: list[RelationComponent] = None, defined_attributes: list[AttributeComponent] = None):
        if defined_attributes is None:
            defined_attributes = []
        self.new_knowledge = new_knowledge if new_knowledge else []
        self.cardinality = cardinality if cardinality else None
        self.requisite = requisite if requisite else RequisiteComponent([])
        self.relations = relations if relations else []
        self.defined_attributes = defined_attributes

    def is_empty(self) -> bool:
        return len(self.new_knowledge) == 0 and len(self.requisite.components) == 0 and len(self.relations) == 0

    def add_requisite(self, requisite: list[Component]):
        self.requisite.add_components(requisite)

    def create_new_signature(self, new_entity: EntityComponent) -> EntityComponent:
        signature: EntityComponent = new_entity.copy()
        linked_entities: list[str] = []  # list of labels already linked
        # add all the keys of all the entities related with the new entity
        for relation in self.relations:
            to_be_related_with = None
            if relation.relation_component_1 == new_entity:
                to_be_related_with = relation.relation_component_2
            elif relation.relation_component_2 == new_entity:
                to_be_related_with = relation.relation_component_1
            if to_be_related_with:

                if len(to_be_related_with.get_keys_and_attributes()) == 1 and \
                        to_be_related_with.label not in linked_entities:
                    for key in to_be_related_with.get_keys():
                        if not new_entity.has_attribute_value(key.value):
                            signature.attributes.append(AttributeComponent(key.get_name(), ValueComponent(Utility.NULL_VALUE),
                                                                           AttributeOrigin(to_be_related_with.get_name(),
                                                                                           key.origin)))
                if to_be_related_with.label not in linked_entities:
                    linked_entities.append(to_be_related_with.label)
                for key in to_be_related_with.get_keys():
                    if key.value != Utility.NULL_VALUE and signature.has_attribute_value(key.value):
                        try:
                            new_entity.get_attributes_by_name_and_origin(key.get_name(), AttributeOrigin(signature.get_name()))[
                                0].origin = key.origin
                            signature.get_attributes_by_name_and_origin(key.get_name(), AttributeOrigin(signature.get_name()))[
                                0].origin = key.origin
                        except:
                            pass
                    try:
                        signature.get_attributes_by_name_and_origin(key.get_name(), key.origin)
                    except AttributeNotFound:
                        try:
                            attribute = signature.get_attributes_by_name(key.get_name())[0]
                            if attribute.value == key.value and attribute.origin.name == signature.get_name() \
                                    and not attribute.origin.origin:
                                new_entity.get_attributes_by_name(key.get_name())[0].origin = key.origin
                                attribute.origin = key.origin
                            else:
                                raise AttributeNotFound('')
                        except AttributeNotFound:
                            signature.attributes.append(AttributeComponent(key.get_name(), ValueComponent(Utility.NULL_VALUE),
                                                                           AttributeOrigin(to_be_related_with.get_name(),
                                                                                           key.origin)))
        signature.attributes.sort(key=lambda x: x.get_name())
        if not signature.attributes:
            signature.attributes.append(
                AttributeComponent(Utility.DEFAULT_ATTRIBUTE, ValueComponent(Utility.NULL_VALUE),
                                   AttributeOrigin(signature.get_name())))
        for attribute in signature.get_keys_and_attributes():
            attribute.value = ValueComponent(Utility.NULL_VALUE)
        return signature

    def convert(self, converter: Converter) -> PropositionConverter:
        return converter.convert_proposition(self)

    def get_entity_by_name(self, name: str):
        for entity in self.get_entities():
            if entity.get_name() == name:
                return entity
        raise EntityNotFound(f'Entity {name} not found.')

    def get_entities(self) -> list[EntityComponent]:
        entities: list[EntityComponent] = []
        for component in self.requisite.components:
            entities += component.get_entities()
        for new_knowledge in self.new_knowledge:
            entities += new_knowledge.get_entities()
        for relation in self.relations:
            entities += relation.get_entities()
        return entities

    def get_entities_to_link_with_new_knowledge(self) -> list[EntityComponent]:
        entities: list[EntityComponent] = []
        for component in self.requisite.components:
            entities += component.get_entities_to_link_with_new_knowledge()
        for new_knowledge in self.new_knowledge:
            entities += new_knowledge.get_entities_to_link_with_new_knowledge()
        for relation in self.relations:
            entities += relation.get_entities_to_link_with_new_knowledge()
        return entities

    def copy(self):
            new_knowledge = [new_knowledge.copy() for new_knowledge in self.new_knowledge]
            cardinality = self.cardinality.copy() if self.cardinality else None
            requisite = self.requisite.copy()
            relations = [relation.copy() for relation in self.relations]
            defined_attributes = [attribute.copy() for attribute in self.defined_attributes]
            return Proposition(new_knowledge, cardinality, requisite, relations, defined_attributes)


class PREFERENCE_PROPOSITION_TYPE(Enum):
    MINIMIZATION = 0
    MAXIMIZATION = 1


class PreferenceProposition(Proposition):
    def __init__(self, requisite: RequisiteComponent = None, relations: list[RelationComponent] = None,
                 weight: str = '1', level: int = 1,
                 discriminant: list[AttributeComponent] = None):
        super().__init__(None, None, requisite, relations)
        if discriminant is None:
            discriminant = []
        self.weight = weight
        self.level = level
        self.discriminant = discriminant
        self.type = PREFERENCE_PROPOSITION_TYPE.MINIMIZATION

    def convert(self, converter: Converter) -> PropositionConverter:
        return converter.convert_preference_proposition(self)

    def add_discriminant(self, discriminant: AttributeComponent):
        if discriminant not in self.discriminant:
            self.discriminant.append(discriminant)

    def copy(self):
        proposition = super().copy()
        preference_proposition = PreferenceProposition(proposition.requisite, proposition.relations, self.weight,
                                                       self.level,
                                                       [discriminant.copy() for discriminant in self.discriminant])
        return preference_proposition



from functools import cmp_to_key

import clingo
from clingo import SolveHandle

from cnl2asp.ASP_elements.solver.clingo_result_parser import ClingoResultParser
from cnl2asp.specification.specification import SpecificationComponent


class TelingoResultParser(ClingoResultParser):

    def __init__(self, specification: SpecificationComponent, model: SolveHandle):
        super().__init__(specification, model)
        self._old_states = []

    def compare(self, item1: str, item2: str):
        if item1 in self._old_states and item2 not in self._old_states:
            return -1
        if item2 in self._old_states and item1 not in self._old_states:
            return 1
        if item1.startswith('There is'):
            return -1
        if item2.startswith('There is'):
            return 1
        return item1 < item2

    def parse_model(self):
        self._get_new_knowledge()
        model = ''
        for state in str(self.model).split("State"):
            if not state.strip():
                continue
            state = state.splitlines()
            model += f"-- In the {state[0].strip().removesuffix(':')} state:\n"
            atoms = []
            for line in state[1:]:
                for elem in line.split(" "):
                    if elem:
                        elem = clingo.parse_term(elem.strip())
                        if elem.name in self.target_predicates:
                            atoms.append(self._clingo_symbol_to_sentence(elem))
            atoms.sort(key=cmp_to_key(self.compare))
            model += '\n'.join(atoms) + '\n'
            self._old_states += atoms
        return model



from __future__ import annotations

from enum import Enum
from functools import total_ordering

from cnl2asp.converter.converter_interface import Converter, OperationConverter
from cnl2asp.specification.attribute_component import ValueComponent
from cnl2asp.specification.component import Component
from cnl2asp.specification.entity_component import EntityComponent


@total_ordering
class Operators(Enum):
    SUM = 0
    DIFFERENCE = 1
    MULTIPLICATION = 2
    DIVISION = 3
    EQUALITY = 4
    INEQUALITY = 5
    GREATER_THAN = 6
    LESS_THAN = 7
    GREATER_THAN_OR_EQUAL_TO = 8
    LESS_THAN_OR_EQUAL_TO = 9
    BETWEEN = 10
    NOTBETWEEN = 11
    ABSOLUTE_VALUE = 12
    CONJUNCTION = 13
    DISJUNCTION = 14
    LEFT_IMPLICATION = 15
    RIGHT_IMPLICATION = 16
    EQUIVALENCE = 17
    NEGATION = 18
    PREVIOUS = 19
    BEFORE = 20
    WEAK_PREVIOUS = 21
    TRIGGER = 22
    ALWAYS_BEFORE = 23
    SINCE = 24
    EVENTUALLY_BEFORE = 25
    PRECEDE = 26
    WEAK_PRECEDE = 27
    NEXT = 28
    AFTER = 29
    WEAK_NEXT = 30
    RELEASE = 31
    ALWAYS_AFTER = 32
    UNTIL = 33
    EVENTUALLY_AFTER = 34
    FOLLOW = 35
    WEAK_FOLLOW = 36

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

class OperationComponent(Component):
    def __init__(self, operator: Operators, *operands: Component):
        self.operation = operator
        self.operands = []
        self.negated = False
        if self.operation == Operators.BETWEEN:
            operands = self.between_operator(operands)
        for operand in operands:
            if not isinstance(operand, Component):
                operand = ValueComponent(operand)
            self.operands.append(operand)


    def between_operator(self, operands):
        operands = list(operands)
        if self.operation == Operators.BETWEEN:
            self.operation = Operators.LESS_THAN_OR_EQUAL_TO
        elif self.operation == Operators.NOTBETWEEN:
            raise NotImplemented("Operator not between not implemented")
        for operand in operands:
            if isinstance(operand, list):
                operands.insert(0, operand[0])
                operands.append(operand[1])
                operands.remove(operand)
        return operands


    def get_entities(self) -> list[EntityComponent]:
        entities = []
        for operand in self.operands:
            entities += operand.get_entities()
        return entities

    def get_entities_to_link_with_new_knowledge(self) -> list[EntityComponent]:
        entities = []
        for operand in self.operands:
            entities += operand.get_entities_to_link_with_new_knowledge()
        return entities

    def convert(self, converter: Converter) -> OperationConverter:
        return converter.convert_operation(self)

    def copy(self):
        operands = [component for component in self.operands]
        return OperationComponent(self.operation, *operands)

    def is_angle(self) -> False:
        for operand in self.operands:
            if operand.is_angle():
                return True
        return False

    def negate(self):
        self.negated = True



from __future__ import annotations

from enum import Enum

from cnl2asp.specification.attribute_component import AttributeComponent
from cnl2asp.specification.component import Component
from cnl2asp.converter.converter_interface import Converter, AggregateConverter
from cnl2asp.specification.entity_component import EntityComponent


class AggregateOperation(Enum):
    COUNT = 0
    SUM = 1
    MAX = 2
    MIN = 3


class AggregateComponent(Component):
    def __init__(self, operation: AggregateOperation, discriminant: list[AttributeComponent], body: list[Component]):
        self.operation = operation
        self.discriminant = discriminant
        self.body = body

    def convert(self, converter: Converter) -> AggregateConverter:
        return converter.convert_aggregate(self)

    def copy(self):
        return AggregateComponent(self.operation, [discriminant.copy() for discriminant in self.discriminant])

    def get_entities(self) -> list[EntityComponent]:
        entities = []
        for entity in self.body:
            entities += entity.get_entities()
        return entities



from __future__ import annotations

import os
import tempfile
import traceback
from enum import Enum
from textwrap import indent
from typing import TextIO

from cnl2asp.utility.utility import Utility
from lark import Lark, UnexpectedCharacters
from lark.exceptions import VisitError

from cnl2asp.ASP_elements.asp_program import ASPProgram
from cnl2asp.converter.cnl2json_converter import Cnl2jsonConverter
from cnl2asp.exception.cnl2asp_exceptions import ParserError
from cnl2asp.specification.attribute_component import AttributeComponent
from cnl2asp.specification.entity_component import EntityComponent
from cnl2asp.converter.asp_converter import ASPConverter
from cnl2asp.parser.parser import CNLTransformer
from cnl2asp.specification.signaturemanager import SignatureManager
from cnl2asp.specification.specification import SpecificationComponent


class SymbolType(Enum):
    DEFAULT = 0
    TEMPORAL = 1


class Symbol:
    def __init__(self, predicate: str, keys: list[str | Symbol], attributes: list[str | Symbol], symbol_type: SymbolType):
        """
        Class for representing the concepts (ASP atoms) structure.

        :param predicate:
        :param keys: the list of attributes that are keys.
        :param attributes: the FULL list (including the keys) of attributes of the atom.
        """
        self.predicate = predicate
        self.keys = keys
        self.attributes = attributes
        self.symbol_type = symbol_type

    def __repr__(self):
        attributes = f'\n{indent(str(self.attributes), "    ")}'
        return f'\n{self.predicate} [{self.symbol_type.name}]: {indent(attributes, "    ")}'


class Cnl2asp:
    def __init__(self, cnl_input: TextIO | str, debug: bool = False):
        self._debug = debug
        if isinstance(cnl_input, str):
            self.cnl_input = cnl_input
            if os.path.isfile(cnl_input):
                self.cnl_input = open(cnl_input).read()
        else:
            self.cnl_input = cnl_input.read()

    def parse_input(self):
        with open(os.path.join(os.path.dirname(__file__), "grammar.lark"), "r") as grammar:
            cnl_parser = Lark(grammar.read(), propagate_positions=True)
            specification: SpecificationComponent = CNLTransformer().transform(cnl_parser.parse(self.cnl_input))
            return specification

    def __is_predicate(self, name: str):
        try:
            SignatureManager.get_signature(name)
            return True
        except:
            return False

    def __get_predicate(self, entity_name: str, attribute: AttributeComponent):
        split_name = attribute.get_name().split('_')
        if len(split_name) > 1:
            for name in split_name:
                if self.__is_predicate(name):
                    return SignatureManager.get_signature(name)
        signature_name = attribute.get_name()
        if attribute.origin != entity_name:
            signature_name = attribute.origin.name
        if self.__is_predicate(signature_name):
            return SignatureManager.get_signature(signature_name)
        return None


    def cnl_to_json(self):
        problem = self.parse_input()
        converter = Cnl2jsonConverter()
        json = problem.convert(converter)
        return json

    def check_syntax(self) -> bool:
        if self.parse_input():
            return True
        return False

    def compile(self, auto_link_entities: bool = True) -> str:
        Utility.AUTO_ENTITY_LINK = auto_link_entities
        try:
            specification: SpecificationComponent = self.parse_input()
        except UnexpectedCharacters as e:
            print(ParserError(e.char, e.line, e.column, e.get_context(self.cnl_input), self.cnl_input.splitlines()[e.line-1], list(e.allowed)))
            return ''
        except VisitError as e:
            print(e.args[0])
            if self._debug:
                traceback.print_exception(e)
            return ''
        try:
            asp_converter: ASPConverter = ASPConverter()
            program: ASPProgram = specification.convert(asp_converter)
            SignatureManager().signatures = []
            return str(program)
        except Exception as e:
            print("Error in asp conversion:", str(e))
            if self._debug:
                traceback.print_exception(e)
            return ''

    def optimize(self, asp_encoding: str):
        from clingo.ast import parse_files
        from ngo import optimize, Predicate
        prg = []
        with tempfile.NamedTemporaryFile(mode="w") as file:
            file.write(asp_encoding)
            file.seek(0)
            parse_files([file.name], prg.append)
            prg = optimize(prg, [Predicate("node",1)], [Predicate("assigned_to",2)])
            optimized_encoding = ''
            for stm in prg:
                optimized_encoding += str(stm) + '\n'
        return optimized_encoding

    def __get_type(self, name: str):
        if SignatureManager.is_temporal_entity(name):
            return SymbolType.TEMPORAL
        return SymbolType.DEFAULT

    def __convert_attribute(self, entity_name: str, attribute: AttributeComponent) -> str | Symbol:
        if attribute.origin and entity_name != attribute.origin.name:
            return Symbol(attribute.origin.name,
                          [self.__convert_attribute(entity_name, AttributeComponent(attribute.get_name(),
                                                                                    attribute.value,
                                                                                    attribute.origin.origin))],
                          [self.__convert_attribute(entity_name, AttributeComponent(attribute.get_name(),
                                                                                    attribute.value,
                                                                                    attribute.origin.origin))],
                          self.__get_type(attribute.origin.name)
                          )
        return attribute.get_name()

    def __convert_signature(self, entity: EntityComponent) -> Symbol:
        keys = []
        attributes = []
        for attribute in entity.get_attributes():
            attributes.append(self.__convert_attribute(entity.get_name(), attribute))
        entity_type = SymbolType.DEFAULT
        if SignatureManager.is_temporal_entity(entity.get_name()):
            entity_type = SymbolType.TEMPORAL
        if entity.get_attributes() != entity.get_keys():
            for key in entity.get_keys():
                keys.append(self.__convert_attribute(entity.get_name(), key))
        return Symbol(entity.get_name(), keys, keys + attributes, entity_type)

    def get_symbols(self) -> list[Symbol]:
        self.compile()
        signatures: list[Symbol] = []
        for signature in SignatureManager().signatures:
            signatures.append(self.__convert_signature(signature))
        SignatureManager().signatures = []
        return signatures



from __future__ import annotations
from typing import Any

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.specification.component import Component
from cnl2asp.specification.constant_component import ConstantComponent
from cnl2asp.specification.problem import Problem


class SpecificationComponent(Component):
    def __init__(self, problems: list[Problem] = None, constants: list[ConstantComponent] = None):
        if problems is None:
            problems = []
        if constants is None:
            constants = []
        self._problems: list[Problem] = problems
        self._constants = constants

    def add_constant(self, constant: ConstantComponent):
        self._constants.append(constant)

    def get_constants(self) -> list[ConstantComponent]:
        return self._constants


    def get_problems(self) -> list[Problem]:
        return self._problems

    def add_problem(self, problem: Problem):
        self._problems.append(problem)

    def convert(self, converter: Converter) -> Any:
        return converter.convert_specification(self)

    def copy(self) -> SpecificationComponent:
        problems = []
        for problem in self._problems:
            problems.append(problem.copy())
        return SpecificationComponent(problems)



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    from cnl2asp.proposition.entity_component import EntityComponent
    from cnl2asp.converter.converter_interface import Converter


class Component(ABC):
    @abstractmethod
    def convert(self, converter: Converter) -> Any:
        """Convert self"""

    @abstractmethod
    def copy(self) -> Any:
        """Return a copy of the component"""

    def get_entities(self) -> list[EntityComponent]:
        return []

    def get_entities_to_link_with_new_knowledge(self) -> list[EntityComponent]:
        return []

    def negate(self):
        pass

    def is_angle(self) -> bool:
        return False




import clingo


class Clingo:
    def __init__(self):
        self.prg = clingo.Control()
        # Clingo configurations
        self.prg.configuration.solve.opt_mode = 'optN'
        self.prg.configuration.solve.models = '1'

    def load(self, encoding: str):
        self.prg.add(encoding)

    def solve(self) -> clingo.SolveHandle:
        self.prg.ground([("base", [])])
        with self.prg.solve(yield_=True) as handle:
            for model in handle:
                if model.optimality_proven:
                    return model
        return model



from __future__ import annotations

from typing import TYPE_CHECKING

from cnl2asp.ASP_elements.asp_operation import ASPOperation
from cnl2asp.specification.attribute_component import AttributeOrigin
from cnl2asp.utility.utility import Utility

from cnl2asp.ASP_elements.asp_element import ASPElement


class ASPValue(ASPElement, str):
    def is_null(self):
        return self == Utility.ASP_NULL_VALUE

    def set_value(self, value: ASPValue):
        return ASPValue(value)


class RangeASPValue(ASPValue):
    def __new__(cls, content):
        return str.__new__(cls, content.replace(' ', '..'))

    def is_null(self):
        if self[0] == Utility.ASP_NULL_VALUE:
            return True
        return False

    def set_value(self, value: ASPValue):
        if Utility.ASP_NULL_VALUE not in self:
            return value
        value = self.replace(Utility.ASP_NULL_VALUE, value).replace('..', ' ')
        return RangeASPValue(value)


class ASPAttribute(ASPElement):
    def __init__(self, name: str, value: ASPValue, origin: AttributeOrigin = None,
                 operations: list[ASPOperation] = None):
        if operations is None:
            operations = []
        self.name = name
        self._value = value
        self.origin = origin
        self.operations = operations

    def is_null(self):
        return self._value.is_null()

    def set_value(self, value: ASPValue):
        self._value = self._value.set_value(value)

    def get_value(self):
        return self._value

    def __str__(self):
        return f'{self._value}'

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, ASPAttribute):
            return False
        return self.name == other.name and self._value == other._value

    def __repr__(self):
        str(self)



from __future__ import annotations

from typing import Any, TYPE_CHECKING

import inflect

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.specification.component import Component
from cnl2asp.specification.name_component import NameComponent

if TYPE_CHECKING:
    from cnl2asp.specification.operation_component import OperationComponent


class ValueComponent(Component, str):
    def convert(self, converter: Converter):
        return converter.convert_value(self)

    def copy(self) -> Any:
        return ValueComponent(self)

    def is_angle(self) -> bool:
        return False


class AngleValueComponent(ValueComponent):

    def copy(self) -> Any:
        return AngleValueComponent(self)

    def is_angle(self) -> bool:
        return True


class RangeValueComponent(ValueComponent):
    def __new__(cls, *content):
        value = content[0]
        if len(content) > 1:
            value += f' {content[1]}'
        return str.__new__(cls, value)

    def convert(self, converter: Converter):
        return converter.convert_range_value(self)

    def copy(self) -> Any:
        return RangeValueComponent(self)


class AttributeOrigin:
    def __init__(self, name: str, origin: AttributeOrigin = None):
        self.name = NameComponent(str(name))
        self.origin = origin
        if origin and self.name == origin.name:
            # avoid nesting same origin
            self.origin = origin.origin

    def is_angle(self):
        if 'angle' in self.name:
            return True
        if self.origin:
            return self.origin.is_angle()
        return False

    def __str__(self):
        if self.origin:
            return str(self.name) + ' ' + str(self.origin)
        return str(self.name)

    def __eq__(self, other):
        if not isinstance(other, AttributeOrigin):
            return False
        if self.name == other.name:
            if self.origin:
                return self.origin == other.origin
            elif not self.origin and not other.origin:
                return True
        return False


def is_same_origin(origin_1: AttributeOrigin, origin_2: AttributeOrigin) -> bool:
    if origin_1 and not origin_2:
        return False
    if origin_2 and not origin_1:
        return False
    # one can be the subset of the other
    if origin_1 == origin_2:
        return True
    if origin_1.origin == origin_2:
        return True
    if origin_2.origin == origin_1:
        return True
    return False


class AttributeComponent(Component):
    def __init__(self, name: str, value: ValueComponent,
                 attribute_origin: AttributeOrigin = None,
                 operations: list[OperationComponent] = None):
        if operations is None:
            operations = []
        self._name = NameComponent(str(name))
        # origin of the attribute, the entity from which it has been taken
        self.origin = attribute_origin
        self.value = AngleValueComponent(value) if self.is_angle() else value
        self.operations = operations

    def name_match(self, name: str) -> bool:
        if self._name == name:
            return True
        return False

    def set_name(self, name: str):
        self._name = NameComponent(name)

    def get_name(self):
        return str(self._name)

    def removesuffix(self, suffix: str):
        self._name.removesuffix(suffix)

    def convert(self, converter: Converter):
        return converter.convert_attribute(self)

    def add_operation(self, operation: OperationComponent):
        self.operations.append(operation)

    def copy(self) -> Any:
        return AttributeComponent(self.get_name(), self.value.copy(), self.origin)

    def is_angle(self) -> bool:
        return 'angle' in self.get_name() or (self.origin.is_angle() if self.origin else False)

    def __eq__(self, other):
        if not isinstance(other, AttributeComponent):
            return False
        return self._name == other._name and self.value == other.value

    def __str__(self):
        return str(self.origin) + ' ' + self.get_name()



import clingo
import telingo
import tempfile

from io import StringIO
import sys

class Telingo:
    def __init__(self):
        self.encoding = ''

    def load(self, encoding: str):
        self.encoding = encoding

    def solve(self):
        with tempfile.NamedTemporaryFile(mode="w") as file:
            file.write(self.encoding)
            file.seek(0)
            sys.stdout = telingo_result = StringIO()
            clingo.clingo_main(telingo.TelApp(), [file.name, "--verbose=0", "--quiet=1,2,2", "--warn=none"])
            sys.stdout = sys.__stdout__
            return telingo_result.getvalue()



from __future__ import annotations

from typing import TYPE_CHECKING

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.specification.component import Component

if TYPE_CHECKING:
    from cnl2asp.specification.entity_component import EntityComponent


class RelationComponent(Component):
    def __init__(self, entity_1: EntityComponent, entity_2: EntityComponent):
        self.relation_component_1 = entity_1
        self.relation_component_2 = entity_2

    def convert(self, converter: Converter):
        converter.convert_relation(self.relation_component_1, self.relation_component_2)

    def copy(self):
        return RelationComponent(self.relation_component_1.copy(), self.relation_component_2.copy())

    def get_entities(self) -> list[EntityComponent]:
        entities: list[EntityComponent] = []
        for entity in self.relation_component_1.get_entities():
            entities += entity.get_entities()
        for entity in self.relation_component_2.get_entities():
            entities += entity.get_entities()
        return entities




from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar, Generic, TYPE_CHECKING

from cnl2asp.specification.component import Component

if TYPE_CHECKING:
    from cnl2asp.specification.entity_component import TemporalEntityComponent
    from cnl2asp.specification.specification import SpecificationComponent
    from cnl2asp.specification.attribute_component import ValueComponent, RangeValueComponent
    from cnl2asp.specification.constant_component import ConstantComponent
    from cnl2asp.specification.problem import Problem
    from cnl2asp.specification.proposition import Proposition, PreferenceProposition, CardinalityComponent
    from cnl2asp.specification.aggregate_component import AggregateComponent
    from cnl2asp.specification.entity_component import EntityComponent
    from cnl2asp.specification.operation_component import OperationComponent
    from cnl2asp.specification.relation_component import RelationComponent
    from cnl2asp.specification.attribute_component import AttributeComponent

ProblemConverter = TypeVar('ProblemConverter')
DomainDefinitionConverter = TypeVar('DomainDefinitionConverter')
PropositionConverter = TypeVar('PropositionConverter')
PreferencePropositionConverter = TypeVar('PreferencePropositionConverter')
NewKnowledgeConverter = TypeVar('NewKnowledgeConverter')
CardinalityConverter = TypeVar('CardinalityConverter')
ConditionConverter = TypeVar('ConditionConverter')
RequisiteConverter = TypeVar('RequisiteConverter')
EntityConverter = TypeVar('EntityConverter')
AggregateConverter = TypeVar('AggregateConverter')
OperationConverter = TypeVar('OperationConverter')
AttributeConverter = TypeVar('AttributeConverter')
ConstantConverter = TypeVar('ConstantConverter')
ValueConverter = TypeVar('ValueConverter')
RelationConverter = TypeVar('RelationConverter')


class Converter(ABC, Generic[ProblemConverter,
                             PropositionConverter, PreferencePropositionConverter,
                             NewKnowledgeConverter, CardinalityConverter,
                             ConditionConverter, RequisiteConverter,
                             EntityConverter, AggregateConverter,
                             OperationConverter, AttributeConverter,
                             ConstantConverter, ValueConverter,
                             RelationConverter]):
    @abstractmethod
    def convert_problem(self, problem: Problem) -> ProblemConverter:
        """Convert problem"""

    @abstractmethod
    def convert_proposition(self, proposition: Proposition) -> PropositionConverter:
        """Convert proposition"""

    @abstractmethod
    def convert_preference_proposition(self, preference: PreferenceProposition) -> PreferencePropositionConverter:
        """Converter preference proposition"""

    @abstractmethod
    def convert_new_knowledge(self, new_knowledge: NewKnowledgeConverter) -> NewKnowledgeConverter:
        """Convert new knowledge"""

    @abstractmethod
    def convert_cardinality(self, cardinality: CardinalityComponent) -> CardinalityConverter:
        """Convert cardinality"""

    @abstractmethod
    def convert_condition(self, condition: ConditionConverter) -> ConditionConverter:
        """Convert condition"""

    @abstractmethod
    def convert_requisite(self, requisite: RequisiteConverter) -> RequisiteConverter:
        """Convert requisite"""

    @abstractmethod
    def convert_entity(self, entity: EntityComponent) -> EntityConverter:
        """Convert entity"""

    @abstractmethod
    def convert_temporal_entity(self, temporal_entity: TemporalEntityComponent):
        "Convert temporal entity"


    @abstractmethod
    def convert_aggregate(self, aggregate: AggregateComponent) -> AggregateConverter:
        """Convert aggregate"""

    @abstractmethod
    def convert_operation(self, operation: OperationComponent) -> OperationConverter:
        """Convert operation"""

    @abstractmethod
    def convert_attribute(self, attribute: AttributeComponent) -> AttributeConverter:
        """Convert attribute"""

    @abstractmethod
    def convert_constant(self, constant: ConstantComponent) -> ConstantConverter:
        """Convert constant"""

    @abstractmethod
    def convert_value(self, value: ValueComponent) -> ValueConverter:
        """Convert value"""

    @abstractmethod
    def convert_range_value(self, value: RangeValueComponent) -> ValueConverter:
        """Convert range value"""

    @abstractmethod
    def convert_relation(self, relation_component_1: Component, relation_component_2: Component) -> RelationConverter:
        """Convert relation"""

    @abstractmethod
    def convert_specification(self, specification: SpecificationComponent):
        """Convert specification"""

    @abstractmethod
    def convert_temporal_entity(self, temporal_entity: TemporalEntityComponent):
        """Convert temporal entity"""




from cnl2asp.ASP_elements.asp_element import ASPElement
from cnl2asp.ASP_elements.asp_program import ASPProgram


class ASPEncoding(ASPElement):
    def __init__(self):
        self._programs: list[ASPProgram] = []
        self._constants: list[(str, str)] = []
        self._telingo_constants: list[str] = ["&true", "&initial", "&false", "&final"]

    def add_constant(self, constant: (str, str)):
        self._constants.append(constant)

    def is_constant(self, var: str) -> bool:
        if var in self._telingo_constants:
            return True
        for constant in self._constants:
            if constant[0] == var:
                return True
        return False

    def add_program(self, program: ASPProgram):
        self._programs.append(program)

    def __str__(self):
        string = ''
        for constant in self._constants:
            if constant[1]:
                string += f'#const {constant[0]} = {constant[1]}.\n'
        for program in self._programs:
            string += str(program)
        return string.strip()



from __future__ import annotations

from typing import TYPE_CHECKING

from cnl2asp.ASP_elements.asp_temporal_formula import ASPTemporalFormula
from cnl2asp.specification.operation_component import Operators
from cnl2asp.ASP_elements.asp_element import ASPElement

if TYPE_CHECKING:
    from cnl2asp.ASP_elements.asp_atom import ASPAtom


class ASPOperation(ASPElement):
    operators = {
        Operators.SUM: '+',
        Operators.DIFFERENCE: '-',
        Operators.MULTIPLICATION: '*',
        Operators.DIVISION: '\\',
        Operators.EQUALITY: '=',
        Operators.INEQUALITY: '!=',
        Operators.GREATER_THAN: '>',
        Operators.LESS_THAN: '<',
        Operators.GREATER_THAN_OR_EQUAL_TO: '>=',
        Operators.LESS_THAN_OR_EQUAL_TO: '<=',
        Operators.ABSOLUTE_VALUE: '|'
    }

    def __init__(self, operator: Operators, *operands: ASPElement):
        self.operator = operator
        self.operands = [operand for operand in operands]

    def get_atom_list(self) -> list[ASPAtom]:
        atom_list: list[ASPAtom] = []
        for operand in self.operands:
            atom_list += operand.get_atom_list()
        return atom_list

    def remove_element(self, element):
        if element in self.operands:
            self.operands.remove(element)
        else:
            for elem in self.operands:
                if element in elem.get_atom_list():
                    elem.remove_element(element)

    def _operator_to_symbol(self, operator):
        return ASPOperation.operators[self.operator]

    def __str__(self) -> str:
        string = ''
        for operand in self.operands:
            if not isinstance(operand, ASPOperation):
                string += str(operand)
            else:
                string += f'({str(operand)})'
            string += f' {self._operator_to_symbol(self.operator)} '
        return string.removesuffix(f' {self._operator_to_symbol(self.operator)} ')

    def __repr__(self):
        return str(self)


class ASPAngleOperation(ASPOperation):
    def __init__(self, operator: Operators, *operands: ASPElement):
        super(ASPAngleOperation, self).__init__(operator, *operands)

    def __str__(self) -> str:
        if self.operator != Operators.SUM and self.operator != Operators.DIFFERENCE and \
                self.operator != Operators.MULTIPLICATION and self.operator != Operators.DIVISION:
            return f' {ASPOperation.operators[self.operator]} '.join(
                ['(' + str(operand) + ')/360' for operand in self.operands])
        else:
            return super(ASPAngleOperation, self).__str__()


class ASPTemporalOperation(ASPOperation):
    asp_temporal_operators = {
        Operators.CONJUNCTION: '&',
        Operators.DISJUNCTION: '|',
        Operators.LEFT_IMPLICATION: '<-',
        Operators.RIGHT_IMPLICATION: '->',
        Operators.EQUIVALENCE: '<>',
        Operators.NEGATION: '~',
        Operators.PREVIOUS: '<',
        Operators.BEFORE: '<',
        Operators.WEAK_PREVIOUS: '<:',
        Operators.TRIGGER: '<*',
        Operators.ALWAYS_BEFORE: '<*',
        Operators.SINCE: '<?',
        Operators.EVENTUALLY_BEFORE: '<?',
        Operators.PRECEDE: '<;',
        Operators.WEAK_PRECEDE: '<:;',
        Operators.NEXT: '>',
        Operators.AFTER: '>',
        Operators.WEAK_NEXT: '>:',
        Operators.RELEASE: '>*',
        Operators.ALWAYS_AFTER: '>*',
        Operators.UNTIL: '>?',
        Operators.EVENTUALLY_AFTER: '>?',
        Operators.FOLLOW: ';>',
        Operators.WEAK_FOLLOW: ';>:',
    }

    def __init__(self, operator: Operators, *operands: ASPElement):
        clean_operands = []
        for operand in operands:
            if isinstance(operand, ASPTemporalFormula):
                clean_operands.append(operand.operations[0])
            else:
                clean_operands.append(operand)

        super(ASPTemporalOperation, self).__init__(operator, *clean_operands)


    def _operator_to_symbol(self, operator):
        return ASPTemporalOperation.asp_temporal_operators.get(operator)

    def __str__(self) -> str:
        if len(self.operands) > 1:
            return f'{super(ASPTemporalOperation, self).__str__()}'
        else:
            string = ''
            string += f'{self._operator_to_symbol(self.operator)} '
            operand = str(self.operands[0])
            if not isinstance(operand, ASPOperation):
                string += operand
            else:
                string += f'({operand})'
            return string


    def temporal_formula_string(self) -> str:
        string = ''
        if len(self.operands) > 1:
            for operand in self.operands:
                operand_to_string = str(operand)
                # replacing atoms starting with initially or finally symbols
                # in the future it might be the case to have different to_string methods
                # based on the context also for the other components
                if operand_to_string.startswith("__"):
                    operand_to_string = operand_to_string.replace("__", ">> ", 1)
                elif operand_to_string.startswith("_"):
                    operand_to_string = operand_to_string.replace("_", "<< ", 1)
                if not isinstance(operand, ASPOperation):
                    string += operand_to_string
                else:
                    string += f'({operand_to_string})'
                string += f' {self._operator_to_symbol(self.operator)} '
            return string.removesuffix(f' {self._operator_to_symbol(self.operator)} ')
        else:
            string = ''
            string += f'{self._operator_to_symbol(self.operator)} '
            operand_to_string = str(self.operands[0])
            # replacing atoms starting with initially or finally symbols
            # in the future it might be the case to have different to_string methods
            # based on the context also for the other components
            if operand_to_string.startswith("__"):
                operand_to_string = operand_to_string.replace("__", ">> ", 1)
            elif operand_to_string.startswith("_"):
                operand_to_string = operand_to_string.replace("_", "<< ", 1)
            if not isinstance(self.operands[0], ASPOperation):
                string += operand_to_string
            else:
                string += f'({operand_to_string})'
            return string


from typing import Any

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.specification.attribute_component import ValueComponent
from cnl2asp.specification.component import Component


class ConstantComponent(Component):
    def __init__(self, name: str, value: ValueComponent):
        self.name = name
        self.value = value

    def convert(self, converter: Converter):
        return converter.convert_constant(self)

    def copy(self) -> Any:
        return ConstantComponent(self.name, self.value.copy())



from cnl2asp.ASP_elements.asp_atom import ASPAtom
from cnl2asp.ASP_elements.asp_element import ASPElement
from cnl2asp.ASP_elements.asp_attribute import ASPAttribute
from cnl2asp.ASP_elements.asp_conjunction import ASPConjunction
from cnl2asp.specification.aggregate_component import AggregateOperation


class ASPAggregate(ASPElement):
    symbols = {
        AggregateOperation.SUM: 'sum',
        AggregateOperation.COUNT: 'count',
        AggregateOperation.MAX: 'max',
        AggregateOperation.MIN: 'min'
    }

    def __init__(self, operation: AggregateOperation, discriminant: list[ASPAttribute], body: ASPConjunction):
        self.operation: AggregateOperation = operation
        self.discriminant = discriminant
        self.body = body

    def get_atom_list(self) -> list[ASPElement]:
        return self.body.get_atom_list()

    def remove_element(self, element: ASPElement):
        self.body.remove_element(element)

    def get_discriminant_attributes_value(self):
        attributes = []
        for elem in self.discriminant:
            if isinstance(elem, ASPAttribute):
                attributes.append(elem.get_value())
            elif isinstance(elem, ASPAtom):
                for attribute in elem.attributes:
                    attributes.append(attribute)
        return attributes

    def __str__(self) -> str:
        return f'#{ASPAggregate.symbols[self.operation]}{{' \
               f'{",".join([str(x) for x in self.discriminant])}' \
               f': {str(self.body)}}}'



from __future__ import annotations

import collections

from cnl2asp.converter.converter_interface import Converter, ProblemConverter
from cnl2asp.specification.constant_component import ConstantComponent
from cnl2asp.specification.entity_component import EntityComponent, TemporalEntityComponent
from cnl2asp.specification.attribute_component import AttributeComponent, RangeValueComponent, is_same_origin
from cnl2asp.specification.problem import Problem
from cnl2asp.specification.proposition import Proposition, NewKnowledgeComponent, ConditionComponent, RequisiteComponent, \
    PreferenceProposition, CardinalityComponent, PREFERENCE_PROPOSITION_TYPE
from cnl2asp.specification.aggregate_component import AggregateComponent
from cnl2asp.specification.operation_component import OperationComponent, Operators
from cnl2asp.specification.relation_component import RelationComponent
from cnl2asp.specification.attribute_component import ValueComponent
from cnl2asp.specification.signaturemanager import SignatureManager
from cnl2asp.specification.specification import SpecificationComponent
from cnl2asp.utility.utility import Utility


class Cnl2jsonConverter(Converter):


    def __init__(self):
        self._json = dict()
        self._json_assignment = collections.defaultdict(set)
        self._json_constant = collections.defaultdict(set)
        self._delayed_operations = []
        self._problem_constants = []

    def convert_specification(self, specification: SpecificationComponent):
        self._json["assignments"] = []
        self._json["constants"] = []
        for constant in specification.get_constants():
            constant.convert(self)
        for problem in specification.get_problems():
            self.convert_problem(problem)
        return self._json

    def convert_problem(self, problem: Problem) -> ProblemConverter:
        for proposition in problem.get_propositions():
            proposition.convert(self)
            self._clear_support_variables()

    def _clear_support_variables(self):
        self._json_assignment = collections.defaultdict(set)
        self._json_constant = collections.defaultdict(set)
        self._delayed_operations = []

    def convert_proposition(self, proposition: Proposition):
        for component in proposition.new_knowledge:
            component.convert(self)
        for component in proposition.requisite.components:
            component.convert(self)
        for delayed_operation in self._delayed_operations:
            self._convert_delayed_operation(delayed_operation)
        for key, value in self._json_assignment.items():
            self._json_assignment.update({key: list(value)})
        for key, value in self._json_constant.items():
            self._json_constant.update({key: list(value)})
        self._json["assignments"].append(dict(self._json_assignment))
        self._json["constants"].append(dict(self._json_constant))

    def __find_attribute(self, attribute_value: ValueComponent):
        for assignment_key, assignment_values in self._json_assignment.items():
            for assignment_value in assignment_values:
                if assignment_value == attribute_value:
                    return assignment_key

    def __is_constant(self, value: ValueComponent):
        for constant in self._problem_constants:
            if constant.name == value:
                return True
        return False

    def _convert_delayed_operation(self, delayed_operation: list[ValueComponent]):
        constants = []
        attributes = []
        for attribute in delayed_operation:
            if self.__is_constant(attribute):
                constants.append(attribute)
            else:
                attributes.append(attribute)
        for attribute in attributes:
            entity = self.__find_attribute(attribute)
            for constant in constants:
                self._json_constant[entity].add(constant)

    def convert_preference_proposition(self, preference: PreferenceProposition):
        self.convert_proposition(preference)

    def convert_new_knowledge(self, head: NewKnowledgeComponent):
        head.new_entity.convert(self)
        for component in head.condition.components:
            component.convert(self)

    def convert_cardinality(self, cardinality: CardinalityComponent) -> (int, int):
        pass

    def convert_condition(self, condition: ConditionComponent):
        for component in condition.components:
            component.convert(self)

    def convert_requisite(self, requisite: RequisiteComponent):
        for component in requisite.components:
            component.convert(self)

    def _is_predicate(self, name: str):
        try:
            SignatureManager.get_signature(name)
            return True
        except:
            return False


    def _get_predicate(self, entity_name: str, attribute: AttributeComponent):
        split_name = attribute.get_name().split('_')
        if len(split_name) > 1:
            for name in split_name:
                if self._is_predicate(name):
                    return SignatureManager.get_signature(name)
        signature_name = attribute.get_name()
        if attribute.origin != entity_name:
            signature_name = attribute.origin.name
        if self._is_predicate(signature_name):
            return SignatureManager.get_signature(signature_name)
        return None

    def convert_entity(self, entity: EntityComponent):
        self._json_assignment[entity.get_name()] = set()
        for attribute in entity.get_keys_and_attributes():
            if attribute.value != Utility.NULL_VALUE:
                predicate = self._get_predicate(entity.get_name(), attribute)
                if predicate:
                    self._json_assignment[predicate.get_name()].add(attribute.value)
                else:
                    self._json_assignment[entity.get_name()].add(attribute.value)

    def convert_temporal_entity(self, temporal_entity: TemporalEntityComponent):
        self.convert_entity(temporal_entity)

    def convert_aggregate(self, aggregate: AggregateComponent):
        for discriminant in aggregate.discriminant:
            discriminant.convert(self)
        for entity in aggregate.body:
            entity.convert(self)

    def _convert_between_operation_without_aggregate(self, operation, operands):
        for operand in operands:
            operand.convert(self)

    def convert_operation(self, operation: OperationComponent):
        delayed_operation = list()
        for operand in operation.operands:
            res = operand.convert(self)
            if res:
                delayed_operation.append(res)
        self._delayed_operations.append(delayed_operation)

    def convert_attribute(self, attribute: AttributeComponent):
        for operation in attribute.operations:
            self.convert_operation(operation)
        return attribute

    def convert_constant(self, constant: ConstantComponent):
        self._problem_constants.append(constant)
        return constant

    def convert_value(self, value: ValueComponent):
        return value

    def convert_range_value(self, value: RangeValueComponent):
        pass

    def convert_relation(self, relation: RelationComponent, new_knowledge: NewKnowledgeComponent = None) -> None:
        pass



from dataclasses import dataclass

import clingo
from clingo import SolveHandle

from cnl2asp.converter.asp_converter import ASPConverter

from cnl2asp.specification.attribute_component import ValueComponent, AttributeComponent

from cnl2asp.specification.entity_component import EntityComponent
from cnl2asp.specification.specification import SpecificationComponent


@dataclass(frozen=True)
class NewKnowledge:
    new_entity: EntityComponent
    subject: list[EntityComponent]
    verb: str
    objects: list[EntityComponent]


class ClingoResultParser:

    def __init__(self, specification: SpecificationComponent, model: SolveHandle):
        self.specification = specification
        self.model = model
        self.target_predicates = []
        self._signatures: list[NewKnowledge] = []

    def _get_new_knowledge(self):
        for programs in self.specification.get_problems():
            for proposition in programs.get_propositions():
                for new_knowledge in proposition.new_knowledge:
                    name = new_knowledge.new_entity.get_name()
                    if name not in self.target_predicates:
                        self.target_predicates.append(new_knowledge.new_entity.get_name())
                        subject = [new_knowledge.subject] if new_knowledge.subject else []
                        self._signatures.append(NewKnowledge(new_knowledge.new_entity.copy(),
                                                             subject, new_knowledge.auxiliary_verb,
                                                             new_knowledge.objects))
                    else:
                        for signature in self._signatures:
                            if signature.new_entity and name == signature.new_entity.get_name():
                                if new_knowledge.subject and new_knowledge.subject.get_name() \
                                        not in [x.get_name() for x in signature.subject]:
                                    signature.subject.append(new_knowledge.subject)

    def _get_signature(self, name: str):
        for signature in self._signatures:
            if name == signature.new_entity.get_name():
                new_entity = signature.new_entity.copy()
                subject = [x.copy() for x in signature.subject.copy()] if signature.subject else signature.subject
                objects = [x.copy() for x in signature.objects] if signature.objects else signature.objects
                return NewKnowledge(new_entity, subject, signature.verb, objects)

    def _entity_printer(self, symbol_name: str, attributes: list[AttributeComponent]):
        attributes_string = ''
        if len(attributes) == 1:
            attributes_string += f'{attributes[0].value}'
        else:
            for attribute in attributes:
                attributes_string += f'with {str(attribute).removeprefix(symbol_name).strip()} equal to {attribute.value}, '
        attributes_string = attributes_string.removesuffix(', ')
        name = ' '.join(symbol_name.split('_'))
        return f"{name} {attributes_string}".strip()

    def convert_attribute_to_entity(self, subject: EntityComponent, atom: EntityComponent):
        res = ''
        if subject:
            attributes_to_print = []
            for attribute in atom.get_keys_and_attributes():
                try:
                    subject_attribute = subject.get_attributes_by_name_and_origin(attribute.get_name(), attribute.origin)[0]
                    subject.get_keys().remove(subject_attribute)
                    attributes_to_print.append(attribute)
                except:
                    pass
            for attribute in attributes_to_print:
                try:
                    atom.attributes.remove(attribute)
                except ValueError:
                    pass
                try:
                    atom.keys.remove(attribute)
                except ValueError:
                    pass
            res = self._entity_printer(subject.get_name(), attributes_to_print)
        return res if res and res != subject.get_name() else "There"

    def _convert_verb(self, verb):
        if verb in ["be ", "be a ", "be an ", "are ", "are a ", "are an ", "is ", "is a ", "is an "]:
            return "is "
        elif verb in ["have ", "have a ", "have an ", "has ", "has a ", "has an "]:
            return "has "
        return ""

    def _parse_clingo_symbol(self, symbol: clingo.Symbol, entity: EntityComponent):
        for idx, attribute in enumerate(entity.get_keys_and_attributes()):
            try:
                attribute.value = ValueComponent(str(symbol.arguments[idx]).removeprefix('\"').removesuffix('\"'))
            except AttributeError:
                attribute.value = ValueComponent(symbol.arguments[idx])

    def _attributes_intersection(self, entity_1: EntityComponent, entity_2: EntityComponent):
        matched_attributes = []
        for attribute in entity_1.get_keys():
            try:
                match = entity_2.get_attributes_by_name_and_origin(attribute.get_name(), attribute.origin)
                matched_attributes += match
            except:
                pass
        return matched_attributes

    def _has_same_value_attribute(self, attribute: AttributeComponent, entity: EntityComponent) -> bool:
        try:
            match = entity.get_attributes_by_name_and_origin(attribute.get_name(), attribute.origin)
            for value in match:
                if attribute.value == value.value:
                    return True
        except:
            return False

    def _check_attributes_list(self, attributes: list[AttributeComponent], entity: EntityComponent) -> bool:
        for attribute in attributes:
            if not self._has_same_value_attribute(attribute, entity):
                return False
        return True

    def _search_entity(self, entity_name: str, attributes: list[AttributeComponent]):
        for programs in self.specification.get_problems():
            for proposition in programs.get_propositions():
                for new_knowledge in proposition.new_knowledge:
                    if new_knowledge.new_entity.get_name() == entity_name:
                        if self._check_attributes_list(attributes, new_knowledge.new_entity):
                            return new_knowledge.new_entity.copy()

    def _convert_subject(self, subject: list[EntityComponent], atom: EntityComponent):
        if len(subject) == 1:
            return self.convert_attribute_to_entity(subject[0], atom)
        elif not subject:
            return self.convert_attribute_to_entity(None, atom)
        elif len(subject) > 1:
            for entity in subject:
                attributes = self._attributes_intersection(entity, atom)
                if self._search_entity(entity.get_name(), attributes):
                    return self.convert_attribute_to_entity(entity, atom)

    def _clingo_symbol_to_sentence(self, symbol: clingo.Symbol):
        signature = self._get_signature(symbol.name)
        self._parse_clingo_symbol(symbol, signature.new_entity)
        string = ''
        subject = self._convert_subject(signature.subject, signature.new_entity)
        if subject != "There":
            verb = self._convert_verb(signature.verb)
        else:
            verb = "is "
        if signature.objects:
            for object_complement in signature.objects:
                string += self.convert_attribute_to_entity(object_complement, signature.new_entity)
        return f"{subject} {verb}" \
               f"{self._entity_printer(signature.new_entity.get_name(), signature.new_entity.get_keys_and_attributes())} " \
               f"{string}".capitalize().strip() + "."

    def parse_model(self):
        self._get_new_knowledge()
        model = ''
        for elem in self.model.symbols(atoms=True):
            if elem.name in self.target_predicates:
                model += self._clingo_symbol_to_sentence(elem) + '\n'
        return model



from cnl2asp.ASP_elements.asp_element import ASPElement
from cnl2asp.ASP_elements.asp_rule import ASPRule


class ASPProgram(ASPElement):
    def __init__(self, name: str = ''):
        self.name: str = name

        self._rules: list[ASPRule] = []

    def add_rule(self, rule: ASPRule):
        self._rules.append(rule)

    def __str__(self) -> str:
        program = f'\n#program {self.name}.\n' if self.name else ''
        for rule in self._rules:
            program += str(rule)
        return program



from __future__ import annotations

from typing import Any

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.specification.component import Component
from cnl2asp.specification.proposition import Proposition


class Problem(Component):
    def __init__(self, name: str = None, propositions: list[Proposition] = None):
        self.name = name
        if propositions is None:
            propositions = []
        self._propositions = propositions


    def add_proposition(self, proposition: Proposition):
        self._propositions.append(proposition)

    def add_propositions(self, propositions: list[Proposition]):
        for proposition in propositions:
            if not proposition.is_empty():
                self._propositions.append(proposition)

    def get_propositions(self) -> list[Proposition]:
        return self._propositions

    def convert(self, converter: Converter) -> Any:
        return converter.convert_problem(self)

    def copy(self):
        propositions = [proposition.copy() for proposition in self._propositions]
        return Problem(self.name, propositions)



from cnl2asp.exception.cnl2asp_exceptions import EntityNotFound, LabelNotFound
from cnl2asp.specification.relation_component import RelationComponent
from cnl2asp.specification.attribute_component import AttributeComponent, ValueComponent, RangeValueComponent
from cnl2asp.specification.component import Component
from cnl2asp.specification.entity_component import EntityComponent
from cnl2asp.specification.proposition import Proposition, NewKnowledgeComponent, CardinalityComponent, \
    PreferenceProposition, PREFERENCE_PROPOSITION_TYPE, ConditionComponent
from cnl2asp.utility.utility import Utility


class PropositionBuilder:

    def __init__(self, proposition: Proposition = None):
        self._original_rule: Proposition = proposition if proposition else Proposition()
        self._derived_rules: list[Proposition] = []

    def get_propositions(self) -> list[Proposition]:
        return [self._original_rule] + self._derived_rules

    def get_cardinality(self):
        return self._original_rule.cardinality

    def get_new_knowledge(self) -> list[NewKnowledgeComponent]:
        return [proposition.new_knowledge for proposition in self.get_propositions()]

    def get_relations(self) -> list[RelationComponent]:
        relations: list[RelationComponent] = []
        for proposition in self.get_propositions():
            for relation in proposition.relations:
                relations.append(relation)
        return relations

    def get_entities(self) -> list[EntityComponent]:
        return self._original_rule.get_entities()

    def get_entity_by_label(self, label: str) -> EntityComponent:
        for entity in self._original_rule.get_entities():
            if entity.label == label:
                return entity
        raise LabelNotFound(f"Label \"{label}\" not declared before.")

    def add_proposition(self, proposition: Proposition):
        self._derived_rules.append(proposition)

    def copy_proposition(self) -> Proposition:
        child = self._original_rule.copy()
        self._derived_rules.append(child)
        return child

    def add_cardinality(self, cardinality: CardinalityComponent):
        for proposition in self.get_propositions():
            proposition.cardinality = cardinality

    def add_requisite(self, component: Component):
        for proposition in self.get_propositions():
            proposition.requisite.components.append(component)

    def add_requisite_list(self, component: list[Component]):
        for proposition in self.get_propositions():
            proposition.requisite.components += component

    def add_new_knowledge(self, new_knowledge: NewKnowledgeComponent):
        for proposition in self.get_propositions():
            proposition.new_knowledge.append(new_knowledge)

    def add_relations(self, relations: list[RelationComponent]):
        for proposition in self.get_propositions():
            proposition.relations += relations

    def duration_clause(self, new_knowledge: NewKnowledgeComponent,
                        duration_value: ValueComponent, attribute_name: str):
        copy: Proposition = self._original_rule.copy()  # create the new proposition to express the duration
        new_knowledge.new_entity.set_name(Utility.create_unique_identifier())  # support entity
        copy.add_requisite([new_knowledge.new_entity])  # add the support entity to the new proposition requisite
        copy.new_knowledge[0].condition = ConditionComponent()
        copy.cardinality = None
        for copy_new_knowledge in copy.new_knowledge:
            # find the starting value of the temporal attribute
            try:
                entity = self._original_rule.get_entity_by_name(attribute_name)
            except EntityNotFound:
                entity = new_knowledge.new_entity
            if entity.is_temporal_entity():
                starting_value = entity.get_temporal_key().value
            else:
                starting_value = entity.get_attributes_by_name(attribute_name)[0].value
            # set the duration to the final entity
            copy_new_knowledge.new_entity.set_attribute_value(attribute_name,
                                                              RangeValueComponent(starting_value,
                                                                                  f'{starting_value}+{duration_value}'))
            copy.relations = []  # set the relations also for the copy
            for relation in self._original_rule.relations:
                for new_knowledge in self._original_rule.new_knowledge:
                    for entity in new_knowledge.condition.components:
                        # do not copy the entity related with the condition part, since
                        # it is not present in the new proposition
                        if not (relation.relation_component_1 == entity or relation.relation_component_2 == entity):
                            copy.relations.append(relation.copy())
            # the final entity is related with all the entities in requisite
            for requisite_entity in copy.requisite.get_entities():
                copy.relations.append(RelationComponent(copy_new_knowledge.new_entity, requisite_entity))
        self._derived_rules.append(copy)

    def create_new_signature(self, new_entity: EntityComponent) -> EntityComponent:
        return self._original_rule.create_new_signature(new_entity)

    def add_discriminant(self, discriminant: list[AttributeComponent]):
        pass

    def add_defined_attributes(self, defined_attributes: list[AttributeComponent]):
        for proposition in self.get_propositions():
            proposition.defined_attributes += defined_attributes

    def add_auxiliary_verb(self, param):
        for new_knowledge in self._original_rule.new_knowledge:
            new_knowledge.auxiliary_verb = param

    def add_subject(self, param):
        for new_knowledge in self._original_rule.new_knowledge:
            new_knowledge.subject = param


class PreferencePropositionBuilder(PropositionBuilder):
    def __init__(self, proposition: PreferenceProposition = None):
        preference_proposition = proposition if proposition else PreferenceProposition()
        super(PreferencePropositionBuilder, self).__init__(preference_proposition)

    def add_weight(self, weight: str):
        for proposition in self.get_propositions():
            proposition.weight = weight

    def add_level(self, level: int):
        for proposition in self.get_propositions():
            proposition.level = level

    def add_discriminant(self, discriminants: list[AttributeComponent]):
        for proposition in self.get_propositions():
            for discriminant in discriminants:
                if discriminant.value != Utility.NULL_VALUE:
                    proposition.add_discriminant(discriminant)

    def add_type(self, proposition_type: PREFERENCE_PROPOSITION_TYPE):
        for proposition in self.get_propositions():
            proposition.type = proposition_type



from typing import Any

import inflect

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.specification.component import Component



class NameComponent(Component):
    def convert(self, converter: Converter) -> Any:
        return self.name

    def copy(self) -> Any:
        return NameComponent(self.name)

    def __init__(self, name: str):
        self.name = name
        self.singular_and_plural_name = [self.__singular(name), self.__plural(name), name]

    def removesuffix(self, suffix):
        self.name = self.name.removesuffix(suffix)
        for value in self.singular_and_plural_name:
            value = value.removesuffix(suffix)

    def __singular(self, string: str):
        if not string:
            return string
        singular = inflect.engine().singular_noun(string)
        return singular if singular else string

    def __plural(self, string: str):
        if not string:
            return string
        plural = inflect.engine().plural(string)
        return plural if plural != self.__singular(string) else string

    def strip(self):
        return self.name.strip()

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, str):
            return other in self.singular_and_plural_name
        if not isinstance(other, NameComponent):
            return False
        return self.name in other.singular_and_plural_name or other.name in self.singular_and_plural_name

    def __contains__(self, item):
        return item in self.name

    def __gt__(self, other):
        if isinstance(other, NameComponent):
            return self.name > other.name



from __future__ import annotations

import re

import inflect

from cnl2asp.parser.parser import CNLTransformer

from cnl2asp.ASP_elements.asp_aggregate import ASPAggregate
from cnl2asp.ASP_elements.asp_atom import ASPAtom
from cnl2asp.ASP_elements.asp_attribute import ASPAttribute, RangeASPValue
from cnl2asp.ASP_elements.asp_conjunction import ASPConjunction
from cnl2asp.ASP_elements.asp_encoding import ASPEncoding
from cnl2asp.ASP_elements.asp_operation import ASPOperation, ASPAngleOperation, ASPTemporalOperation
from cnl2asp.ASP_elements.asp_program import ASPProgram
from cnl2asp.ASP_elements.asp_rule import ASPRule, ASPRuleHead, ASPWeakConstraint
from cnl2asp.ASP_elements.asp_attribute import ASPValue
from cnl2asp.ASP_elements.asp_temporal_formula import ASPTemporalFormula

from cnl2asp.converter.converter_interface import Converter
from cnl2asp.exception.cnl2asp_exceptions import AttributeNotFound, EntityNotFound, CompilationError

from cnl2asp.specification.constant_component import ConstantComponent
from cnl2asp.specification.entity_component import EntityComponent, TemporalEntityComponent
from cnl2asp.specification.attribute_component import AttributeComponent, RangeValueComponent, is_same_origin
from cnl2asp.specification.component import Component
from cnl2asp.specification.problem import Problem
from cnl2asp.specification.proposition import Proposition, NewKnowledgeComponent, ConditionComponent, RequisiteComponent, \
    PreferenceProposition, CardinalityComponent, PREFERENCE_PROPOSITION_TYPE
from cnl2asp.specification.aggregate_component import AggregateComponent
from cnl2asp.specification.operation_component import OperationComponent, Operators
from cnl2asp.specification.relation_component import RelationComponent
from cnl2asp.specification.attribute_component import ValueComponent
from cnl2asp.specification.signaturemanager import SignatureManager
from cnl2asp.specification.specification import SpecificationComponent
from cnl2asp.utility.utility import Utility


class EntityToAtom:
    def __init__(self, entity: EntityComponent, atom: ASPAtom):
        self.entity = entity
        self.atom = atom

def is_arithmetic_operator(operator):
    if operator <= Operators.DIVISION:
        return True
    return False

operators_negation = {
    Operators.EQUALITY: Operators.INEQUALITY,
    Operators.INEQUALITY: Operators.EQUALITY,
    Operators.GREATER_THAN: Operators.LESS_THAN_OR_EQUAL_TO,
    Operators.LESS_THAN: Operators.GREATER_THAN_OR_EQUAL_TO,
    Operators.GREATER_THAN_OR_EQUAL_TO: Operators.LESS_THAN,
    Operators.LESS_THAN_OR_EQUAL_TO: Operators.GREATER_THAN,
    Operators.BETWEEN: Operators.NOTBETWEEN
}

class ForbiddenLink:
    def __init__(self, entity_1: EntityComponent, entity_2: EntityComponent, attribute_1: AttributeComponent,
                 attribute_2: AttributeComponent):
        self.entity_1 = entity_1
        self.entity_2 = entity_2
        self.attribute_1 = attribute_1
        self.attribute_2 = attribute_2

    def contains(self, atom_1, atom_2, attribute):
        if atom_1 == self.atom_1 or atom_1 == self.atom_2:
            if atom_2 == self.atom_1 or atom_2 == self.atom_2:
                if attribute == self.attribute_1 or self.attribute_2:
                    return True
        return False


class ASPConverter(Converter[ASPProgram,
                             ASPRule, ASPWeakConstraint,
                             ASPRuleHead, (int, int),
                             ASPConjunction, ASPConjunction,
                             ASPAtom, ASPAggregate,
                             ASPOperation, ASPAttribute,
                             None, ASPValue, None]):

    def __init__(self):
        self._asp_encoding: ASPEncoding = ASPEncoding()
        self._program: ASPProgram = ASPProgram()
        self._atoms_in_current_rule: list[EntityToAtom] = []  # variable used to track the conversion entity -> atom
        self._created_fields: list[str] = []
        self._aggregates: list[ASPAggregate] = []
        self._operations: list[ASPOperation] = []
        self._forbidden_links: list[ForbiddenLink] = []
        self._converted_complex_entities: list[str] = [] # name of complex entities already converted, used to track if their values have been already converted.

    def get_trailing_number(self, s: str):
        m = re.search(r'\d+$', s)
        return int(m.group()) if m else None

    def _get_initialized_attributes(self, proposition: Proposition):
        for entity in proposition.get_entities():
            for attribute in entity.get_keys_and_attributes():
                if attribute.value != Utility.NULL_VALUE:
                    self._created_fields.append(attribute.value)

    def create_new_field_value(self, name: str) -> ASPValue:
        result = re.sub(r'[AEIOU]', '', name, flags=re.IGNORECASE).upper()
        if result in self._created_fields:
            trailing_number = self.get_trailing_number(result)
            if trailing_number:
                new_value = int(trailing_number) + 1
                result = result.replace(str(trailing_number), str(new_value))
            else:
                result += '1'
            result = self.create_new_field_value(result)
        self._created_fields.append(result)
        return result

    def convert_specification(self, specification: SpecificationComponent):
        for constant in specification.get_constants():
            constant.convert(self)
        for problem in specification.get_problems():
            self._asp_encoding.add_program(problem.convert(self))
            self._program = ASPProgram()
        return self._asp_encoding

    def convert_problem(self, problem: Problem) -> ASPProgram:
        self._program.name = problem.name
        for proposition in problem.get_propositions():
            self._program.add_rule(proposition.convert(self))
            self.clear_support_variables()
        return self._program

    def clear_support_variables(self):
        self._atoms_in_current_rule = []
        self._created_fields = []
        self._forbidden_links = []
        self._aggregates = []
        self._operations = []

    def convert_proposition(self, proposition: Proposition) -> ASPRule:
        # Create a new signature for each new entity
        self._get_initialized_attributes(proposition)
        head = []
        cardinality = None
        body = None
        self._created_fields = proposition.defined_attributes
        if proposition.new_knowledge:
            for new_knowledge in proposition.new_knowledge:
                self._forbidden_links += self.forbidden_links(new_knowledge, proposition.requisite)
                new_knowledge = new_knowledge.convert(self)
                self.add_atoms_operations(new_knowledge.condition)
                head.append(new_knowledge)
        if proposition.cardinality:
            cardinality = proposition.cardinality.convert(self)
        if proposition.requisite:
            body: ASPConjunction = proposition.requisite.convert(self)
            self.add_atoms_operations(body)
        if proposition.relations:
            for relation in proposition.relations:
                relation.convert(self)
        self.__move_operations(body)
        return ASPRule(body, head, cardinality)

    def add_atoms_operations(self, context):
        atoms_operations = []
        for atom in context.get_atom_list():
            for attribute in atom.attributes:
                for operation in attribute.operations:
                    if str(operation) not in atoms_operations:
                        atoms_operations.append(str(operation))
                        context.add_element(operation)

    def __move_operations(self, body):
        for operation in self._operations:
            for operand in operation.operands:
                for aggregate in self._aggregates:
                    if operand in aggregate.get_discriminant_attributes_value() and not aggregate in operation.operands:
                        aggregate.body.add_element(operation)
                        body.remove_element(operation)

    def convert_preference_proposition(self, preference: PreferenceProposition) -> ASPWeakConstraint:
        rule = self.convert_proposition(preference)
        discriminant = [attribute.convert(self) for attribute in preference.discriminant if
                        CNLTransformer.is_variable(attribute.value)]
        weight = preference.weight
        if preference.type == PREFERENCE_PROPOSITION_TYPE.MAXIMIZATION:
            weight = '-' + preference.weight.upper()
        weak_constraint = ASPWeakConstraint(rule.body, weight,
                                            preference.level, discriminant)
        return weak_constraint

    def convert_new_knowledge(self, head: NewKnowledgeComponent) -> ASPRuleHead:
        new_knowledge = head.new_entity.convert(self)
        condition = head.condition.convert(self) if head.condition.components else None
        return ASPRuleHead(new_knowledge, condition)

    def convert_cardinality(self, cardinality: CardinalityComponent) -> (int, int):
        return cardinality.lower_bound, cardinality.upper_bound

    def convert_condition(self, condition: ConditionComponent) -> ASPConjunction:
        return self.__create_conjunction(condition.components)

    def convert_requisite(self, requisite: RequisiteComponent) -> ASPConjunction:
        return self.__create_conjunction(requisite.components)

    def __create_conjunction(self, components: list[Component]) -> ASPConjunction:
        asp_conjunction = ASPConjunction([])
        for component in components:
            asp_conjunction.add_element(component.convert(self))
        return asp_conjunction

    def convert_entity(self, entity: EntityComponent) -> ASPAtom:
        atom = ASPAtom(entity.get_name(), [attribute.convert(self) for attribute in entity.keys + entity.attributes],
                       entity.negated, entity.is_before, entity.is_after, entity.is_initial, entity.is_final)
        self._atoms_in_current_rule.append(EntityToAtom(entity, atom))
        return atom

    def convert_temporal_entity(self, temporal_entity: TemporalEntityComponent):
        if temporal_entity.get_name() not in self._converted_complex_entities:
            for value, idx in temporal_entity.values.items():
                self._program.add_rule(ASPRule(head=[ASPRuleHead(ASPAtom(temporal_entity.get_name(),
                                                                       [ASPAttribute(temporal_entity.get_name().removesuffix('s'), ASPValue(idx)),
                                                                        ASPAttribute('value', ASPValue(f'\"{value}\"'))]))]))
            self._converted_complex_entities.append(temporal_entity.get_name())
        return self.convert_entity(temporal_entity)

    def __has_single_key(self, entity: EntityComponent) -> bool:
        entity_keys = entity.get_keys()
        return len(entity_keys) == 1 and \
               entity.get_attributes_by_name_and_origin(entity_keys[0].get_name(), entity_keys[0].origin)[0].value == Utility.ASP_NULL_VALUE

    def _match_discriminant_attributes_with_body(self, discriminant: list, body: ASPConjunction):
        unmatched_discriminant_attributes = []
        # create a variable to match discriminant attribute with the atoms attributes with the same name
        for attribute in discriminant:
            discriminant_value = attribute.get_value()
            attributes_to_be_equal_discriminant_value = [attribute]
            attribute_matched = False
            for atom in body.get_atom_list():
                atom_matched_attributes = atom.get_attributes_list(attribute.name, attribute.origin)
                if len(atom_matched_attributes) > 1:
                    # TODO give the aggregate in the warning
                    print(f"Warning: multiple attribute with same name for atom {atom}, "
                          f"the first has been taken for aggregate.")
                if atom_matched_attributes:
                    attribute_matched = True
                    atom_matched_attribute = atom_matched_attributes[0]
                    if not atom_matched_attribute.is_null():
                        if not discriminant_value.is_null() and discriminant_value != atom_matched_attribute.value:
                            continue
                        discriminant_value = atom_matched_attribute.get_value()
                    attributes_to_be_equal_discriminant_value.append(atom_matched_attribute)
            if not attribute_matched:
                unmatched_discriminant_attributes.append(attribute)
                try:
                    discriminant += [attribute.convert(self) for attribute in
                                     SignatureManager.get_signature(attribute.name).get_keys()]
                except EntityNotFound:
                    raise Exception(f"Impossible to use attribute \"{attribute.origin} {attribute}\" in aggregate.")
            discriminant = [attribute for attribute in discriminant if
                            attribute not in unmatched_discriminant_attributes]
            if discriminant_value.is_null():
                discriminant_value = ASPValue(self.create_new_field_value(attribute.name))
            for attribute_to_set in attributes_to_be_equal_discriminant_value:
                attribute_to_set.set_value(discriminant_value)
        return discriminant

    def _match_discriminant_atom_with_body(self, discriminant: list[ASPAtom], body: ASPConjunction):
        for discriminant_elem in discriminant:
            for body_elem in body.get_atom_list():
                if discriminant_elem == body_elem:
                    for attribute in discriminant_elem.attributes:
                        self._link_atom_to_attribute(body_elem, attribute, discriminant_elem, [])

    def convert_aggregate(self, aggregate: AggregateComponent) -> ASPAggregate:
        # discriminant can be an atom or an attribute
        discriminant = [discr.convert(self) for discr in aggregate.discriminant]
        body = ASPConjunction([component.convert(self) for component in aggregate.body])
        if [x for x in discriminant if isinstance(x, ASPAttribute)]:
            discriminant = self._match_discriminant_attributes_with_body(discriminant, body)
        else:
            self._match_discriminant_atom_with_body(discriminant, body)
        aggregate = ASPAggregate(aggregate.operation, discriminant, body)
        self._aggregates.append(aggregate)
        return aggregate

    def _is_list_of_aggregates(self, operands) -> bool:
        for operand in operands:
            if not isinstance(operand, ASPAggregate):
                return False
        return True

    def _convert_operation_of_list_of_aggregate(self, operation, operands):
        operations: list[ASPOperation] = []
        fields: list[ASPValue] = []
        for operand in operands:
            new_field = ASPValue(self.create_new_field_value(operand.operation.name))
            operations.append(ASPOperation(Operators.EQUALITY, operand, new_field))
            fields.append(new_field)
        for i, field_1 in enumerate(fields):
            for j, field_2 in enumerate(fields[i + 1:]):
                operations.append(ASPOperation(operation.operation, field_1, field_2))
        for operation in operations:
            self._operations.append(operation)
        return operations

    def _convert_between_operation_without_aggregate(self, operation, operands):
        operation1 = ASPOperation(operation.operation, operands[0], operands[1])
        operation2 = ASPOperation(operation.operation, operands[1], operands[2])
        self._operations.append(operation1)
        self._operations.append(operation2)
        return ASPConjunction([operation1, operation2])

    def convert_operation(self, operation: OperationComponent) -> ASPOperation | [ASPOperation]:
        operands = []
        if operation.negated and operation.operation < Operators.CONJUNCTION:
            operation.operation = operators_negation[operation.operation]
        is_operation_on_angle = False
        for operand in operation.operands:
            if operand.is_angle():
                is_operation_on_angle = True
            operands.append(operand.convert(self))
        if is_operation_on_angle:
            return ASPAngleOperation(operation.operation, *operands)
        if self._is_list_of_aggregates(operands):
            return self._convert_operation_of_list_of_aggregate(operation, operands)
        if not is_arithmetic_operator(operation.operation) and len(operands) == 3 \
                and not isinstance(operands[1], ASPAggregate) and operation.operation < Operators.CONJUNCTION:
            return self._convert_between_operation_without_aggregate(operation, operands)
        if operation.operation >= Operators.CONJUNCTION:
            return ASPTemporalFormula([ASPTemporalOperation(operation.operation, *operands)], operation.negated)
        operation = ASPOperation(operation.operation, *operands)
        self._operations.append(operation)
        return operation

    def convert_attribute(self, attribute: AttributeComponent) -> ASPAttribute:
        attribute_name = inflect.engine().singular_noun(attribute.get_name()) if inflect.engine().singular_noun(attribute.get_name()) else attribute.get_name()
        operations = [operation.convert(self) for operation in attribute.operations]
        return ASPAttribute(attribute_name, attribute.value.convert(self), attribute.origin, operations)

    def convert_constant(self, constant: ConstantComponent) -> None:
        if constant.value and constant.value.isalpha():
            value = f'"{constant.value}"'
        else:
            value = constant.value.convert(self)
        self._asp_encoding.add_constant((constant.name, value))

    def convert_value(self, value: ValueComponent) -> ASPValue:
        if not value:
            return ASPValue(value)
        if not self._asp_encoding.is_constant(value) and not value == Utility.NULL_VALUE and not value.isnumeric() and not value.isupper():
            return ASPValue(f'"{value}"')
        return ASPValue(value)

    def convert_range_value(self, value: RangeValueComponent) -> ASPValue:
        return RangeASPValue(value)

    def forbidden_links(self, new_knowledge: NewKnowledgeComponent, requisite: RequisiteComponent) -> list[
        ForbiddenLink]:
        new_knowledge_links: list[AttributeComponent] = []
        for new_entity in new_knowledge.new_entity.get_entities():
            for condition in new_knowledge.condition.get_entities():
                for condition_key in condition.get_keys():
                    try:
                        attributes = new_entity.get_attributes_by_name(condition_key.get_name())
                        for attribute in attributes:
                            if is_same_origin(attribute.origin, condition_key.origin):
                                new_knowledge_links.append(attribute)
                    except AttributeNotFound:
                        pass
        forbidden_links: list[ForbiddenLink] = []
        for attribute in new_knowledge_links:
            for requisite_entity in requisite.get_entities():
                try:
                    requisite_entity_attributes = requisite_entity.get_attributes_by_name(attribute.get_name())
                    for requisite_entity_attribute in requisite_entity_attributes:
                        if is_same_origin(attribute.origin, requisite_entity_attribute.origin):
                            forbidden_links.append(
                                ForbiddenLink(new_knowledge.new_entity, requisite_entity,
                                              attribute.convert(self), requisite_entity_attribute.convert(self)))
                except AttributeNotFound:
                    pass
        return forbidden_links

    def convert_relation(self, relation_component_1: Component, relation_component_2: Component, new_knowledge: NewKnowledgeComponent = None) -> None:
        relation_entities_1 = relation_component_1.get_entities()
        if len(relation_entities_1) > 1:
            for entities in relation_entities_1:
                self.convert_relation(entities, relation_component_2)
        else:
            relation_component_1 = relation_entities_1[0]
            relation_entities_2 = relation_component_2.get_entities()
            if len(relation_entities_2) > 1:
                for entities in relation_entities_2:
                    self.convert_relation(relation_component_1, entities)
            else:
                self.entity_to_atoms_and_link(relation_component_1, relation_entities_2[0])


    def entity_to_atoms_and_link(self, entity_1: EntityComponent, entity_2: EntityComponent, new_knowledge: NewKnowledgeComponent = None):
        atom_1 = None
        atom_2 = None
        for pair in self._atoms_in_current_rule:
            # look for the exactly same entity
            if pair.entity is entity_1:
                atom_1 = pair.atom
            if pair.entity is entity_2:
                atom_2 = pair.atom
        # if entity not found, search for a copy
        if not atom_1:
            for pair in self._atoms_in_current_rule:
                if pair.entity == entity_1:
                    atom_1 = pair.atom
        if not atom_2:
            for pair in self._atoms_in_current_rule:
                if pair.entity == entity_2:
                    atom_2 = pair.atom
        if not atom_1 or not atom_2:
            raise Exception(f'Relation between {entity_1.get_name()} and {entity_2.get_name()} not found.')
        forbidden_links = self.get_forbidden_link(entity_1, entity_2)
        self._link_two_atoms(entity_1, entity_2, atom_1, atom_2, forbidden_links)

    def get_forbidden_link(self, entity_1, entity_2) -> list[ASPAttribute]:
        forbidden_attributes = []
        for forbidden_link in self._forbidden_links:
            if (entity_1 == forbidden_link.entity_1 or entity_1 == forbidden_link.entity_2) and \
                    (entity_2 == forbidden_link.entity_1 or entity_2 == forbidden_link.entity_2):
                forbidden_attributes += [forbidden_link.attribute_1, forbidden_link.attribute_2]
        return forbidden_attributes

    def _link_two_atoms(self, entity_1: EntityComponent, entity_2: EntityComponent, atom_1: ASPAtom, atom_2: ASPAtom,
                        forbidden_links: list[ASPAttribute] = None):
        if forbidden_links is None:
            forbidden_links = []
        linked_attributes: list[ASPAttribute] = forbidden_links  # list of already linked attributes
        for key in entity_1.get_keys():
            atom_1_keys = atom_1.get_attributes_list_by_name_and_origin(key.get_name(), key.origin)
            for atom_1_key in atom_1_keys:
                self._link_atom_to_attribute(atom_2, atom_1_key, atom_1, linked_attributes)

        for key in entity_2.get_keys():
            atom_2_keys = atom_2.get_attributes_list_by_name_and_origin(key.get_name(), key.origin)
            for atom_2_key in atom_2_keys:
                if atom_2_key in linked_attributes:
                    continue
                self._link_atom_to_attribute(atom_1, atom_2_key, atom_2, linked_attributes)

    def _link_atom_to_attribute(self, atom_1: ASPAtom,
                                attribute: ASPAttribute, atom_2: ASPAtom, linked_attributes: list[ASPAttribute]):
        atom_1_attributes = atom_1.get_attributes_list(attribute.name)
        for atom_1_attribute in atom_1_attributes:
            if (not atom_1_attribute.is_null() and atom_2.has_attribute(atom_1_attribute)) or (not attribute.is_null() and atom_1.has_attribute(attribute)):
                continue
            if atom_1_attribute in linked_attributes:
                continue
            if is_same_origin(atom_1_attribute.origin, attribute.origin):
                # if both attributes have a value, skip it. The next attribute should be the correct one
                new_var = ASPValue(self.create_new_field_value('_'.join([atom_1.name, attribute.name])))
                if not attribute.is_null() and not atom_1_attribute.is_null():
                    continue
                if not attribute.is_null():
                    new_var = attribute.get_value()
                if not atom_1_attribute.is_null():
                    new_var = atom_1_attribute.get_value()
                atom_1.set_attributes_value([ASPAttribute(attribute.name, new_var, attribute.origin)])
                attribute.set_value(new_var)
                # add the two linked attribute keys to the already linked attributes list
                linked_attributes.append(attribute)
                linked_attributes.append(atom_1_attribute)
                # we found a match we can skip. We do not want to link also the next value, this is
                # the case in which the entity has more attributes matching
                break



from __future__ import annotations

import abc
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, TYPE_CHECKING

from cnl2asp.exception.cnl2asp_exceptions import AttributeNotFound
from cnl2asp.converter.converter_interface import Converter, EntityConverter
from cnl2asp.specification.attribute_component import AttributeComponent, ValueComponent, \
    AttributeOrigin, is_same_origin
from cnl2asp.specification.component import Component
from cnl2asp.specification.name_component import NameComponent
from cnl2asp.specification.relation_component import RelationComponent
from cnl2asp.utility.utility import Utility

if TYPE_CHECKING:
    from cnl2asp.parser.proposition_builder import PropositionBuilder


class EntityType(Enum):
    GENERIC = 0
    TIME = 1
    DATE = 2
    STEP = 3
    ANGLE = 4
    SET = 5
    LIST = 6


class EntityComponent(Component):
    def __init__(self, name: str, label: str, keys: list[AttributeComponent], attributes: list[AttributeComponent],
                 negated: bool = False, entity_type: EntityType = EntityType.GENERIC,
                 is_before: bool = False, is_after: bool = False, is_initial: bool = False,
                 is_final: bool = False, auxiliary_verb: str = ""):
        self._name = NameComponent(name)
        self.label = label
        self.keys = keys if keys else []
        self.attributes = attributes if attributes else []
        for attribute in self.attributes:
            if attribute.origin is None:
                attribute.origin = AttributeOrigin(self.get_name())
        self.negated = negated
        self.entity_type = entity_type
        self.auxiliary_verb = auxiliary_verb
        self.is_before = is_before
        self.is_after = is_after
        self.is_initial = is_initial
        self.is_final = is_final

    def label_is_key_value(self):
        if self.label and len(self.get_keys()) == 1 and self.get_keys()[0].value == Utility.NULL_VALUE:
            return True
        return False

    def set_label_as_key_value(self):
        self.set_attributes_value([AttributeComponent(self.get_keys()[0].get_name(),
                                   ValueComponent(self.label), self.get_keys()[0].origin)])

    def set_name(self, name: str):
        self._name = NameComponent(name)

    def get_name(self):
        return str(self._name)

    def is_initialized(self) -> bool:
        if self.get_keys_and_attributes():
            return True
        return False

    def has_attribute_value(self, value: ValueComponent) -> bool:
        for attribute in self.attributes:
            if attribute.value == value:
                return True
        return False

    def convert(self, converter: Converter) -> EntityConverter:
        return converter.convert_entity(self)

    def negate(self):
        self.negated = not self.negated

    def get_entity_identifier(self):
        return self._name

    def get_keys(self) -> list[AttributeComponent]:
        return self.keys if self.keys else self.attributes

    def get_attributes(self) -> list[AttributeComponent]:
        return self.attributes

    def get_keys_and_attributes(self) -> list[AttributeComponent]:
        return self.keys + self.attributes

    def set_attributes_value(self, attributes: list[AttributeComponent], proposition: PropositionBuilder = None):
        for attribute in attributes:
            if isinstance(attribute, EntityComponent):
                if not proposition:
                    raise RuntimeError('Error in compilation phase.')
                proposition.add_relations([RelationComponent(self, attribute)])
            else:
                if attribute.origin:
                    entity_attributes = self.get_attributes_by_name_and_origin(attribute.get_name(), attribute.origin)
                else:
                    origin = AttributeOrigin(str(self.get_name()))
                    entity_attributes = self.get_attributes_by_name_and_origin(attribute.get_name(), origin)
                for entity_attribute in entity_attributes:
                    entity_attribute.value = attribute.value
                    entity_attribute.operations = attribute.operations

    def set_attribute_value(self, attribute_name: str, value: ValueComponent, origin: AttributeOrigin = None):
        if not origin:
            origin = self.get_attributes_by_name(attribute_name)[0].origin
        self.set_attributes_value([AttributeComponent(attribute_name, value, origin)])

    def get_attributes_by_name_and_origin(self, name: str, origin: AttributeOrigin = None) -> list[AttributeComponent]:
        attributes = []
        origin = AttributeOrigin(self.get_name()) if not origin else origin
        for attribute in self.get_keys_and_attributes():
            if attribute.name_match(name) and is_same_origin(attribute.origin, origin):
                attributes.append(attribute)
        if attributes:
            return attributes
        else:
            error_msg = f'Entity \"{self.get_name()}\" do not contain attribute \"{origin} {name}\".'
            try:
                hint = self.get_attributes_by_name(name)
                error_msg += f'\nDid you mean \"{hint[0]}\"?'
            except AttributeNotFound:
                pass
            raise AttributeNotFound(error_msg)

    def get_attributes_by_name(self, name: str) -> list[AttributeComponent]:
        attributes = []
        for attribute in self.get_keys_and_attributes():
            if attribute.name_match(name):
                attributes.append(attribute)
        if attributes:
            return attributes
        raise AttributeNotFound(f'Entity \"{self.get_name()}\" do not contain attribute \"{name}\".')

    def copy(self) -> EntityComponent:
        keys = [key.copy() for key in self.keys]
        attributes = [attribute.copy() for attribute in self.attributes]
        return EntityComponent(self.get_name(), self.label, keys,
                               attributes, self.negated, self.entity_type,
                               self.is_before, self.is_after, self.is_initial, self.is_final,
                               self.auxiliary_verb)

    def get_entities(self) -> list[EntityComponent]:
        return [self]

    def get_entities_to_link_with_new_knowledge(self) -> list[EntityComponent]:
        return [self]

    def is_temporal_entity(self) -> bool:
        return False

    def __eq__(self, other):
        if not isinstance(other, EntityComponent):
            return False
        return self._name == other._name \
               and self.label == other.label \
               and self.keys == other.keys \
               and self.attributes == other.attributes \
               and self.negated == other.negated \
               and self.entity_type == other.entity_type


class TemporalEntityComponent(EntityComponent):
    def __init__(self, name: str, label: str,
                 lhs_range_value: ValueComponent,
                 rhs_range_value: ValueComponent, step: ValueComponent,
                 entity_type: EntityType, values: dict = None):
        super().__init__(name, label,
                         [AttributeComponent(name, ValueComponent('_'), AttributeOrigin(name))],
                         [AttributeComponent('value', ValueComponent('_'), AttributeOrigin(name))],
                         entity_type=entity_type)
        step = 1 if not step else step
        self.values: dict = self._compute_values(lhs_range_value, rhs_range_value, step) if not values else values

    def get_temporal_value_id(self, value: ValueComponent):
        temporal_value_id = self.values.get(value)
        if temporal_value_id is not None:
            return temporal_value_id
        raise KeyError(f'Value "{value}" out of "{self.get_name()}" range')

    def _compute_values(self, lhs_value: ValueComponent, rhs_value: ValueComponent,
                        step: ValueComponent) -> dict:
        """
        Compute all the values in the range and assign a label starting from 0
        :param lhs_value: left hand side value of the range
        :param rhs_value: right hand side value of the range
        :param step: step of increment from lhs to rhs
        :return: a dict with each value associated to a number
        """
        elements = dict()
        counter = 1
        try:
            if self.entity_type == EntityType.TIME:
                start = datetime.strptime(lhs_value, '%I:%M %p')
                end = datetime.strptime(rhs_value, '%I:%M %p')
                elements[start.strftime('%I:%M %p')] = 0
                start = start + timedelta(minutes=int(step))
                while start <= end:
                    if start > end:
                        start = end
                    elements[start.strftime('%I:%M %p')] = counter
                    counter += 1
                    start = start + timedelta(minutes=int(step))
            elif self.entity_type == EntityType.DATE:
                start = datetime.strptime(lhs_value, '%d/%m/%Y')
                end = datetime.strptime(rhs_value, '%d/%m/%Y')
                elements[start.strftime('%d/%m/%Y')] = 0
                start = start + timedelta(days=int(step))
                while start <= end:
                    if start > end:
                        start = end
                    elements[start.strftime('%d/%m/%Y')] = counter
                    counter += 1
                    start = start + timedelta(days=int(step))
            elif self.entity_type == EntityType.STEP:
                start = lhs_value
                end = rhs_value
                elements[int(start)] = 0
                for i in range(int(start) + 1, int(end) + 1):
                    elements[i] = counter
                    counter += 1
        except:
            raise Exception(f"Entity {self.get_name()}, entity type do not match {lhs_value} and {rhs_value}")
        return elements

    def is_temporal_entity(self) -> bool:
        return True

    def get_temporal_key(self) -> AttributeComponent:
        return self.keys[0]

    def copy(self) -> Any:
        copy = TemporalEntityComponent(self.get_name(), self.label, None,
                                       None, None, self.entity_type, self.values)
        copy.set_attributes_value(super(TemporalEntityComponent, self).copy().attributes)
        copy.set_attributes_value(super(TemporalEntityComponent, self).copy().keys)
        return copy

    def convert(self, converter: Converter):
        return converter.convert_temporal_entity(self)


class ComplexEntityComponent(EntityComponent):
    def __init__(self, name: str, complex_entity_id: ValueComponent, keys: list[AttributeComponent],
                 attributes: list[AttributeComponent], entity_type: EntityType, values: list[ValueComponent] = None):
        if values is None:
            self.values = []
        else:
            self.values = values
        self.entity_identifier = complex_entity_id
        super().__init__(name, '', keys, attributes, False, entity_type)

    def get_entity_identifier(self):
        return self.entity_identifier

    def set_values(self, values):
        self.values = values


class SetEntityComponent(ComplexEntityComponent):
    def __init__(self, set_id: ValueComponent = None, values: list[ValueComponent] = None):
        super().__init__('set', set_id, [AttributeComponent('set_id', set_id, AttributeOrigin('set')),
                                         AttributeComponent('element', ValueComponent(Utility.NULL_VALUE),
                                                            AttributeOrigin('set'))],

                         [], EntityType.SET, values)

    def is_value_in_set(self, value: ValueComponent) -> bool:
        for set_value in self.values:
            if value == set_value:
                return True
        return False

    def set_label_as_key_value(self):
        if self.label in self.values:
            self.set_attribute_value('element', ValueComponent(self.label))

    def copy(self) -> SetEntityComponent:
        return SetEntityComponent(self.entity_identifier, self.values)


class ListEntityComponent(ComplexEntityComponent):
    def __init__(self, list_id: ValueComponent = None, values: list[ValueComponent] = None):
        super().__init__('list', list_id, [AttributeComponent('list_id', list_id, AttributeOrigin('list')),
                                           AttributeComponent('index', ValueComponent(Utility.NULL_VALUE),
                                                              AttributeOrigin('list'))],
                         [AttributeComponent('element', ValueComponent(Utility.NULL_VALUE), AttributeOrigin('list'))],
                         EntityType.LIST, values)

    def set_label_as_key_value(self):
        if self.label in self.values:
            self.set_attribute_value('element', ValueComponent(self.label))
            self.set_attribute_value('index', ValueComponent(self.values.index(ValueComponent(self.label))))

    def copy(self) -> ListEntityComponent:
        return ListEntityComponent(self.entity_identifier, self.values)

    def set_shifted_value(self, shift_value: int, starting_value: str, element_variable: ValueComponent):
        element_index = (self.values.index(ValueComponent(starting_value)) + shift_value) % len(self.values)
        element = self.values[element_index]
        if element_variable:
            element_index = element_variable
        self.set_attribute_value('index', ValueComponent(element_index))
        self.set_attribute_value('element', ValueComponent(element))

    def set_index_value(self, index: int, element_variable: ValueComponent):
        element = self.values[index]
        if element_variable:
            index = element_variable
        self.set_attribute_value('index', ValueComponent(index))
        self.set_attribute_value('element', ValueComponent(element))



from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cnl2asp.ASP_elements.asp_atom import ASPAtom


class ASPElement:
    def get_attributes_list(self, name):
        raise Exception("Looking for attribute in a non-ASPAtom element.")

    def get_atom_list(self) -> list[ASPAtom]:
        return []

    def remove_element(self, element):
        pass

    def is_null(self):
        return False



from textwrap import dedent

from setuptools import setup, find_packages

setup(
    name='cnl2asp',
    version='1.3.1',
    description='A tool for converting a Controlled Natural Language based on English into Answer Set Programming',
    long_description=dedent('''\
                            A tool for converting a Controlled Natural Language based on English into Answer Set Programming
                            
                            Documentation is available at:
                            https://dodaro.github.io/cnl2asp/
                            '''),
    url='https://github.com/dodaro/cnl2asp/tree/main',
    license='Apache License',
    author='Carmine Dodaro',
    author_email='carmine.dodaro@unical.it',
    maintainer='Simone Caruso',
    maintainer_email='simone.caruso@edu.unige.it',
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['tests*']),
    package_data={'': ['grammar.lark']},
    install_requires=['lark', 'inflect', 'multipledispatch'],
    python_requires=">=3.10"
)


from __future__ import annotations
from multipledispatch import dispatch

from cnl2asp.ASP_elements.asp_element import ASPElement
from cnl2asp.specification.signaturemanager import SignatureManager


class ASPConjunction(ASPElement):
    def __init__(self, conjunction: list[ASPElement]):
        self.conjunction = conjunction

    def remove_null_atoms(self):
        for atom in self.conjunction:
            if atom.is_null():
                self.remove_element(atom)
        for elem in self.conjunction:
            for atom in elem.get_atom_list():
                if atom.is_null():
                    elem.remove_element(atom)

    @dispatch(list)
    def add_element(self, element: list[ASPElement]):
        self.conjunction += element

    @dispatch(ASPElement)
    def add_element(self, element: ASPElement):
        if isinstance(element, ASPConjunction):
            for elem in element.conjunction:
                self.add_element(elem)
        else:
            self.conjunction.append(element)

    def remove_element(self, element: ASPElement):
        self.conjunction.remove(element)

    def get_atom_list(self) -> list[ASPElement]:
        atom_list = []
        for x in self.conjunction:
            atom_list += x.get_atom_list()
        return atom_list

    def __str__(self) -> str:
        return ', '.join([str(x) for x in self.conjunction if str(x)])

    def __eq__(self, other):
        if not isinstance(other, ASPConjunction):
            return False
        return self.conjunction == other.conjunction

    def __repr__(self):
        return str(self)
